#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  8 10:38:30 2023

@author: alex
"""

import numpy as np

from pyro.dynamic import plane

from pyro.control import controller




###############################################################################

class PLaneController( controller.StaticController ) :
    
    ############################
    def __init__( self  ):
        """ """
        
        # Dimensions
        self.k   = 1 
        self.m   = 2
        self.p   = 6 
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Plane Controller'
        
        self.v_ref  = 40
        self.y_ref  = 0.0
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        v     = y[3]
        theta = y[2]
        y     = y[1]
        
        T         = +10 * ( self.v_ref - v )
        
        theta_ref = +0.1 * ( self.y_ref - y )
        
        delta     = -0.5 * ( theta_ref - theta )
        
        u = np.array([ T , delta ])

        
        return u


sys = plane.Plane2D()
sys.l_w = 0.0

sys.x0   = np.array([0,0,0.2,15,0,0])

ctl = PLaneController()

cl_sys = ctl + sys

cl_sys.compute_trajectory( 2 )

cl_sys.plot_trajectory()
cl_sys.animate_simulation( time_factor_video = 0.2 )






# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.control  import nonlinear
from pyro.control  import robotcontrollers
from pyro.planning import plan
from pyro.analysis import simulation
###############################################################################

sys  = pendulum.SinglePendulum()

###############################################################################

# Planning

traj   = plan.Trajectory.load('rrt.npy')
q_goal = np.array([-3.14])

###############################################################################

# P
kp = 5
kd = 0
ki = 0
p_ctl      = robotcontrollers.JointPID( 1 , kp , ki, kd)
p_ctl.rbar = q_goal

# PD
kp = 5
kd = 2
ki = 0
pd_ctl      = robotcontrollers.JointPID( 1 , kp , ki, kd)
pd_ctl.rbar = q_goal

# PID
kp = 5
kd = 2
ki = 1
pid_ctl      = robotcontrollers.JointPID( 1 , kp , ki, kd)
pid_ctl.rbar = q_goal

# Computed Torque
ctc_ctl      = nonlinear.ComputedTorqueController( sys )
ctc_ctl.rbar = q_goal
ctc_ctl.w0   = 2.0
ctc_ctl.zeta = 0.8 

# Sliding Mode 
sld_ctl      = nonlinear.SlidingModeController( sys )
sld_ctl.lam  = 1
sld_ctl.gain = 5
sld_ctl.rbar = q_goal

# OpenLoop with traj
traj_ctl = plan.OpenLoopController( traj )

# Computed Torque with traj
traj_ctc_ctl      = nonlinear.ComputedTorqueController( sys , traj )
traj_ctc_ctl.rbar = q_goal
traj_ctc_ctl.w0   = 2.0
traj_ctc_ctl.zeta = 0.8 

# Sliding Mode with traj
traj_sld_ctl      = nonlinear.SlidingModeController( sys , traj )
traj_sld_ctl.lam  = 1
traj_sld_ctl.gain = 5
traj_sld_ctl.rbar = q_goal

###############################################################################

# Controller selection

ctl = p_ctl
#ctl = pd_ctl
#ctl = pid_ctl
#ctl = ctc_ctl
#ctl = sld_ctl
#ctl = traj_ctl
#ctl = traj_ctc_ctl 
#ctl = traj_sld_ctl


###############################################################################

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
q0 = 0
tf = 10
cl_sys.x0[0] = q0
cl_sys.compute_trajectory(tf,10001,'euler')
cl_sys.plot_trajectory('xu')
sys.plot_phase_plane()
cl_sys.plot_phase_plane_trajectory_closed_loop()
cl_sys.animate_simulation()




# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

# Target
q_desired = np.array([0.5,0.5])
r_desired = torque_controlled_robot.forward_kinematic_effector( q_desired )


# Effector PD 

model = torque_controlled_robot

effector_pd      = robotcontrollers.EndEffectorPD( model )
effector_pd.rbar = r_desired
effector_pd.kp   = np.array([100, 100 ])
effector_pd.kd   = np.array([  0,   0 ])

# Closed-loops

robot_with_effector_pd = effector_pd + torque_controlled_robot 

# Simulations

tf = 4
robot_with_effector_pd.x0 = np.array([0,0,0,0])
robot_with_effector_pd.compute_trajectory( tf )
robot_with_effector_pd.plot_trajectory('xu')
robot_with_effector_pd.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

from pyro.dynamic                 import pendulum
from pyro.dynamic.tranferfunction import ss2tf
from pyro.dynamic.statespace      import linearize
    
non_linear_sys = pendulum.SinglePendulum()

non_linear_sys.lc1     = 1
non_linear_sys.m1      = 2
non_linear_sys.I1      = 2
non_linear_sys.d1      = 2
non_linear_sys.gravity = 9.81

linearized_sys = linearize( non_linear_sys )
siso_sys       = ss2tf( linearized_sys, 0, 0)

print('Poles',siso_sys.poles)
print('num',siso_sys.num)
print('den',siso_sys.den)

siso_sys.bode_plot()

# Shortcut
non_linear_sys.plot_linearized_bode()


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

import numpy as np

from pyro.dynamic import tranferfunction
from pyro.dynamic import statespace
from pyro.dynamic import massspringdamper
from pyro.control import linear


# Plant
m = 1
k = 1
b = 0

plant = massspringdamper.SingleMass(m,k,b)
tf    = tranferfunction.ss2tf( plant, 0, 0)

tf.bode_plot()
tf.pz_map()


#Controller
kp  = 5
ki  = 2
kd  = 4
tau = 0.1

pid = linear.PIDController(kp,ki,kd,tau)

pid.rbar[0] = 1 # set point


closed_loop_sys = pid + plant

cl_ss = statespace.linearize( closed_loop_sys )
cl_tf = tranferfunction.ss2tf( cl_ss, 0, 0)

cl_tf.bode_plot()
cl_tf.pz_map()

closed_loop_sys.plot_trajectory('y')
closed_loop_sys.plot_trajectory('u')
closed_loop_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import pendulum
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = pendulum.SinglePendulum()

sys.x_ub = np.array([+10, +10])
sys.x_lb = np.array([-10,  -10])

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [501,501] , [21] )

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ -3.14 , 0 ]) # target
qcf.INF  = 500

qcf.R[0,0] = 1.0

qcf.S[0,0] = 10.0
qcf.S[1,1] = 10.0


# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)
#dp = dprog.DynamicProgramming2DRectBivariateSpline(grid_sys, qcf)

#dp.solve_bellman_equation( animate_cost2go = True )

#dp.compute_steps(200)
# dp.plot_policy()

#dp.solve_bellman_equation( tol = 1)
dp.solve_bellman_equation( tol = 0.1 , animate_cost2go = True )
#dp.solve_bellman_equation( tol = 1 , animate_policy = True )
#dp.plot_cost2go(150)

#dp.animate_cost2go( show = False , save = True )
#dp.animate_policy( show = False , save = True )

dp.clean_infeasible_set()
dp.plot_cost2go_3D()
dp.plot_policy()

ctl = dp.get_lookup_table_controller()


#ctl.plot_control_law( sys = sys , n = 100)


#asign controller
cl_sys = ctl + sys
cl_sys.x0   = np.array([0., 0.])
cl_sys.compute_trajectory( 10, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.control  import nonlinear
from pyro.analysis import simulation
###############################################################################

sys  = pendulum.SinglePendulum()


# Controller

traj = simulation.Trajectory.load('pendulum_rrt.npy')

ctl  = nonlinear.SlidingModeController( sys , traj )

ctl.lam  = 5
ctl.gain = 2

# goal
ctl.rbar = np.array([-3.14])

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = np.array([0.1,0])
cl_sys.compute_trajectory(tf=5, n=10001, solver='euler') 
# Note: Use "euler" solver when using sliding mode controllers
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep 29 15:03:52 2023

@author: alex
"""

##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic  import mountaincar
from pyro.planning import trajectoryoptimisation
from pyro.analysis import costfunction
##############################################################################

sys  = mountaincar.MountainCar()

sys.x_ub = np.array([+0.2,+2.0])
sys.x_lb = np.array([-1.7,-2.0])

sys.u_ub[0] = +1.5
sys.u_lb[0] = -1.5


# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ 0 , 0 ]) # target
qcf.INF  = 30

qcf.R[0,0] = 10.0

qcf.S[0,0] = 10.0
qcf.S[1,1] = 10.0

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( sys , 0.1, 40 )

planner.x_start = np.array([-1.0,+0.0])
planner.x_goal  = np.array([+0.0,+0.0])

planner.init_dynamic_plot()
planner.set_linear_initial_guest()
planner.compute_optimal_trajectory()
# planner.show_solution()
planner.animate_solution()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 25 12:16:59 2019

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import vehicle_steering
###############################################################################

# Vehicule dynamical system
#sys = vehicle.KinematicCarModel()
sys = vehicle_steering.KinematicCarModelwithObstacles()

# Set default wheel velocity and steering angle
sys.ubar = np.array([2,0.2])
sys.x0   = np.array([0,0,0])

# Plot open-loop behavior
tf = 100
sys.compute_trajectory( tf )
sys.plot_trajectory()

# Animate the simulation
sys.animate_simulation( time_factor_video = 5 )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 29 16:30:18 2022

@author: alex
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import cartpole
from pyro.planning import trajectoryoptimisation
from pyro.analysis import simulation
###############################################################################


sys  = cartpole.UnderActuatedRotatingCartPole()

sys.cost_function.Q[0,0] = 1.0
sys.cost_function.Q[1,1] = 1.0
sys.cost_function.Q[2,2] = 100
sys.cost_function.Q[3,3] = 1.0

#Max/Min torque
sys.u_ub[0] = +20
sys.u_lb[0] = -20

## Planning

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( sys , 0.1 , 30 )

# Load an initial guess found with RRT
init_traj = simulation.Trajectory.load('underactuatedcartpole_rrt.npy')
planner.set_initial_trajectory_guest( init_traj )

planner.x_start = np.array([0,-3.14,0,0])
planner.x_goal  = np.array([0,0,0,0])

planner.maxiter = 500
planner.compute_optimal_trajectory()
planner.show_solution()
planner.animate_solution( is_3d = True ) 


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""

import numpy as np

from pyro.dynamic  import vehicle_steering
from pyro.planning import discretizer
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming

sys  = vehicle_steering.HolonomicMobileRobotwithObstacles()

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , (51,51) , (3,3) ) 

# Cost Function
cf = costfunction.QuadraticCostFunction.from_sys( sys )
cf.INF = 1500

# VI algo
vi = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys , cf )

vi.solve_bellman_equation()

# Closed-loop Law

vi.clean_infeasible_set()
vi.plot_cost2go_3D()
#vi.plot_policy(0)
#vi.plot_policy(1)

ctl = vi.get_lookup_table_controller()





# Closed loop
cl_sys = ctl + sys

# Simulation and animation
cl_sys.x0   = np.array([9,0])
cl_sys.compute_trajectory(tf=20)
#cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 13 13:37:57 2024

@author: alex
"""

import gymnasium as gym

from stable_baselines3 import PPO
from stable_baselines3.common.env_util import make_vec_env

# Parallel environments
vec_env = make_vec_env("Pendulum-v1", n_envs=1)

model = PPO("MlpPolicy", vec_env, verbose=1)
model.learn(total_timesteps=250000)

obs = vec_env.reset()

while True:
    action, _states = model.predict(obs)
    obs, rewards, dones, info = vec_env.step(action)
    vec_env.render("human")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct  3 08:27:06 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.rocket        import Rocket
from pyro.analysis.costfunction import QuadraticCostFunction
from pyro.dynamic.statespace    import linearize
from pyro.control.lqr           import synthesize_lqr_controller


# Non-linear model
sys = Rocket()

sys.inertia = 400

sys.xbar =  np.array([0,2.2,0,0,0,0])
sys.ubar =  np.array([1,0]) * sys.mass * sys.gravity # Nominal trust = weight
    
# Linear model
ss  = linearize( sys , 0.01 )

# Cost function
cf  = QuadraticCostFunction.from_sys( sys )
cf.Q[0,0] = 1
cf.Q[1,1] = 10000
cf.Q[2,2] = 0.1
cf.Q[3,3] = 0
cf.Q[4,4] = 10000
cf.Q[5,5] = 0
cf.R[0,0] = 0.01
cf.R[1,1] = 10.0

# LQR controller
ctl = synthesize_lqr_controller( ss , cf , sys.xbar , sys.ubar )

ctl.ubar

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys
cl_sys.x0 = np.array([1,10,-0.8,0,0,0])
cl_sys.compute_trajectory(10)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Tue Oct 23 20:45:37 2018

@author: Alexandre
"""



###############################################################################
import numpy as np
##############################################################################
from pyro.dynamic import system
from pyro.dynamic import vehicle_steering
###############################################################################


# NOTE: models in development and not fully validated


###############################################################################
        
class LateralDynamicBicycleModelwithSpeedInput( vehicle_steering.KinematicCarModel ):
    """ 
    Equations of Motion
    -------------------------
    dv_y    = (F_yf*cos(delta)+F_yr)/m - v_x*dtheta         "lateral force sum (F_y = m*dv_y) gives lateral acceleration ddv_y"
    ddtheta = (a*F_yf*cos(delta)-b*F_yr)/Iz                 "Torque sum at mass center (T = I*ddtheta) gives the angular acceleration of the vehicle ddtheta"
    dtheta  = dtheta                                        "dtheta is already a state which is the yaw rate"
    dX      = v_x*cos(theta)-v_y*sin(theta)                 "To obtain cartesian position"
    dY      = v_x*sin(theta)+v_y*cos(theta)

    Where 
    F_yf is the lateral force applied perpendicularly with the front wheel (N)
    F_yr is the lateral force applied perpendicularly with the rear wheel (N)
    v_y  is the lateral velocity of the mass center (m/s)
    v_x  is the longitudinal velocity of the mass center (m/s)
    delta is the steering angle (rad)
    theta is the yaw angle of the vehicle (rad)
    m     is the mass of the vehicle (kg)
    Iz    is the inertia for a rotation along the z-axis (kg*m^2)
    a     is the distance from the front axle to the mass centre (m)
    b     is the distance from the rear axle to the mass centre (m)
    (X,Y) is the cartesian position of the vehicle
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 5   
        self.m = 2   
        self.p = 5
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Lateral Dynamic Bicyle Model'
        self.state_label = ['v_y','dtheta','theta','X','Y']
        self.input_label = ['delta', 'v_x']
        self.output_label = ['v_y','dtheta','theta','X','Y']
        
        # Units
        self.state_units = ['[m/s]','[rad/s]','[rad]','[m]','[m]']
        self.input_units = ['[rad]', '[m/sec]']
        self.output_units = ['[m/s]','[rad/s]','[rad]','[m]','[m]']

        
        # Model param (used for animation purposes only change the values)
        self.width  = 2.00
        self.a      = 2.00
        self.b      = 3.00
        self.lenght = self.a+self.b    
        self.lenght_tire = 0.40
        self.width_tire = 0.15
        
        self.g = 9.81
        self.mass = 2000
        self.Iz = 1.00/12.00*self.mass*((self.a+self.b)**2+self.width**2) 
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = self.width * 2
    
    #######################################
    def piecewise_tiremodel(self,x,u):
        
        #Tire-road friction coefficient
        self.mu = 0.9
        #Compute normal forces on the tires
        F_nf = self.mass*self.g*self.b/(self.b+self.a)
        F_nr = self.mass*self.g*self.a/(self.b+self.a)
        #Compute the max forces available
        max_F_f = F_nf*self.mu
        max_F_r = F_nr*self.mu
        #Compute the lateral "slip-slope"
        self.max_alpha_stat = 0.12
        slip_ratio_f = max_F_f/(self.max_alpha_stat)
        slip_ratio_r = max_F_r/(self.max_alpha_stat)
        if (u[1] == 0):
            slip_f = 0
            slip_r = 0
            F_yf   = 0
            F_yr   = 0
        else:
            slip_f = u[0]-np.arctan((x[0]+self.a*x[1])/u[1])
            slip_r = np.arctan((self.b*x[1]-x[0])/u[1])
            
        if (slip_f<-0.12):
            F_yf = -max_F_f
        elif (slip_f > 0.12):
            F_yf = max_F_f
        else:
            F_yf = slip_ratio_f*slip_f
            
        if (slip_r<-0.12):
            F_yr = -max_F_r
        elif (slip_r > 0.12):
            F_yr = max_F_r
        else:
            F_yr = slip_ratio_r*slip_r

        return F_yf,F_yr   
        
    #############################
        
    def f(self, x = np.zeros(5) , u = np.zeros(2) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        F_yf,F_yr = self.piecewise_tiremodel(x,u)
        
        dx[0] = (F_yf*np.cos(u[0])+F_yr)/self.mass-u[1]*x[1]
        dx[1] = (self.a*F_yf*np.cos(u[0])-self.b*F_yr)/self.Iz
        dx[2] = x[1]
        dx[3] = u[1]*np.cos(x[2])-x[0]*np.sin(x[2])
        dx[4] = u[1]*np.sin(x[2])-x[0]*np.cos(x[2])
        
        return dx
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q =  np.array( [ x[3] ,  x[4] , x[2] , u[0] ] )
        
        return q
    
    
    
##############################################################################
        
class LateralDynamicBicycleModelwithForceInputs( LateralDynamicBicycleModelwithSpeedInput ):
    
    """ 
    Equations of Motion
    -------------------------
    
    dx   = V cos ( phi )
    dy   = V sin ( phi )
    dphi = V/l tan ( beta )
    
    STATES_DERIVATIVE
    dv_x    = (F_xf*cos(delta)-F_yf*sin(delta)+F_xr)/m-v_y*dtheta   "longitudinal force sum (F_x = m*dv_x) gives longitudinal acceleration dv_x"
    dv_y    = (F_yf*cos(delta)+F_xf*sin(delta)+F_yr)/m - v_x*dtheta "lateral force sum (F_y = m*dv_y) gives lateral acceleration dv_y"
    ddtheta = (a*(F_yf*cos(delta)+F_xf*sin(delta))-b*F_yr)/Iz       "Torque sum at mass center (T = I*ddtheta) gives the angular acceleration of the vehicle ddtheta"
    dtheta  = dtheta                                                "dtheta is already a state which is the yaw rate"
    dX      = v_x*cos(theta)-v_y*sin(theta)                         "To obtain cartesian position"
    dY      = v_x*sin(theta)+v_y*cos(theta)
    
    INPUTS 
    delta is the steering angle (rad)
    F_xf is the longitudinal force given by the front propulsion (N)
    F_xr is the longitudinal force given by the rear propulsion (N)
    
    
    Where 
    F_yf is the lateral force applied perpendicularly with the front wheel (N)
    F_yr is the lateral force applied perpendicularly with the rear wheel (N)
    v_y  is the lateral velocity of the mass center (m/s)
    v_x  is the longitudinal velocity of the mass center (m/s)
    delta is the steering angle (rad)
    theta is the yaw angle of the vehicle (rad)
    m     is the mass of the vehicle (kg)
    Iz    is the inertia for a rotation along the z-axis (kg*m^2)
    a     is the distance from the front axle to the mass centre (m)
    b     is the distance from the rear axle to the mass centre (m)
    (X,Y) is the cartesian position of the vehicle
    
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 6   
        self.m = 3   
        self.p = 6
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Full Dynamic Bicyle Model without slip'
        self.state_label = ['v_x','v_y','dtheta','theta','X','Y']
        self.input_label = ['delta', 'F_xf','F_xr']
        self.output_label = ['v_x','v_y','dtheta','theta','X','Y']
        
        # Units
        self.state_units = ['[m/s]','[m/s]','[rad/s]','[rad]','[m]','[m]']
        self.input_units = ['[rad]', '[N]','[N]']
        self.output_units = ['[m/s]','[m/s]','[rad/s]','[rad]','[m]','[m]']

        
        # Model param (used for animation purposes only change the values)
        self.width  = 2.00
        self.a      = 2.00
        self.b      = 3.00
        self.lenght = self.a+self.b    
        self.lenght_tire = 0.40
        self.width_tire = 0.15
        
        self.g = 9.81
        self.mass = 1000
        self.Iz = 1.00/12.00*self.mass*((self.a+self.b)**2+self.width**2) 
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = 5
        
    #############################
    def piecewise_tiremodel(self,x,u):
        
        #Tire-road friction coefficient
        self.mu = 0.9
        F_nf = self.mass*self.g*self.b/(self.b+self.a)
        F_nr = self.mass*self.g*self.a/(self.b+self.a)
        #Compute the max forces available
        max_F_f = F_nf*self.mu
        max_F_r = F_nr*self.mu
        #Compute the lateral "slip-slope"
        self.max_alpha_stat = 0.12
        slip_ratio_f = max_F_f/(self.max_alpha_stat)
        slip_ratio_r = max_F_r/(self.max_alpha_stat)
        if x[0] == 0:
            slip_f = 0
            slip_r = 0
        elif(x[0]>0):
            slip_f = np.arctan((x[1]+self.a*x[2])/x[0])-u[0]
            slip_r = np.arctan((x[1]-self.b*x[2])/x[0])
        else:
            slip_f = -(np.arctan((x[1]+self.a*x[2])/x[0])+u[0])
            slip_r = -np.arctan((x[1]-self.b*x[2])/x[0])            
        if (slip_f<-0.12):
            F_yf = max_F_f
        elif (slip_f > 0.12):
            F_yf = -max_F_f
        else:
            F_yf = -slip_ratio_f*slip_f
            
        if (slip_r<-0.12):
            F_yr = max_F_r
        elif (slip_r > 0.12):
            F_yr = -max_F_r
        else:
            F_yr = -slip_ratio_r*slip_r
        
        return F_yf,F_yr 
        
    #############################
        
    def f(self, x = np.zeros(6) , u = np.zeros(3) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        dx = np.zeros(self.n) # State derivative vector
        F_yf,F_yr = self.piecewise_tiremodel(x,u)
        
        dx[0] = (u[1]*np.cos(u[0])-F_yf*np.sin(u[0])+u[2])/self.mass-x[1]*x[2]
        dx[1] = (F_yf*np.cos(u[0])+u[1]*np.sin(u[0])+F_yr)/self.mass-x[0]*x[2]
        dx[2] = (self.a*(F_yf*np.cos(u[0])+u[1]*np.sin(u[0]))-self.b*F_yr)/self.Iz
        dx[3] = x[2]
        dx[4] = x[0]*np.cos(x[3])-x[1]*np.sin(x[3])
        dx[5] = x[0]*np.sin(x[3])+x[1]*np.cos(x[3])
        
        return dx
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q =  np.array( [ x[4] ,  x[5] , x[3] , u[0] ] )
        
        return q
    
    
    
##############################################################################
        
class FullDynamicBicycleModelwithTorquesInputs( LateralDynamicBicycleModelwithSpeedInput ):
    
    """ 
    Equations of Motion
    -------------------------
    STATES_DERIVATIVE
    dv_x    = (F_xf*cos(delta)-F_yf*sin(delta)+F_xr)/m-v_y*dtheta   "longitudinal force sum (F_x = m*dv_x) gives longitudinal acceleration dv_x"
    dv_y    = (F_yf*cos(delta)+F_xf*sin(delta)+F_yr)/m - v_x*dtheta "lateral force sum (F_y = m*dv_y) gives lateral acceleration dv_y"
    ddtheta = (a*(F_yf*cos(delta)+F_xf*sin(delta))-b*F_yr)/Iz       "Torque sum at mass center (T = I*ddtheta) gives the angular acceleration of the vehicle ddtheta"
    dtheta  = dtheta                                                "dtheta is already a state which is the yaw rate"
    dX      = v_x*cos(theta)-v_y*sin(theta)                         "To obtain cartesian position"
    dY      = v_x*sin(theta)+v_y*cos(theta)
    domega_f = (T_F-F_xf*R)/I_propf
    domega_r = (T_r-F_xr*R)/I_propr
    
    ***** DEPENDS DU MODELE DE TIRE COMPLET EX: S=omega/v_x et F_x,F_y = f(F_z,alpha,S et mu) *****
    
    INPUTS 
    delta is the steering angle (rad)
    T_f is the front wheel torque (N)
    T_r is the rear wheel torque (N)
    
    
    Where 
    F_yf is the lateral force applied perpendicularly with the front wheel (N)
    F_yr is the lateral force applied perpendicularly with the rear wheel (N)
    v_y  is the lateral velocity of the mass center (m/s)
    v_x  is the longitudinal velocity of the mass center (m/s)
    delta is the steering angle (rad)
    theta is the yaw angle of the vehicle (rad)
    m     is the mass of the vehicle (kg)
    Iz    is the inertia for a rotation along the z-axis (kg*m^2)
    a     is the distance from the front axle to the mass centre (m)
    b     is the distance from the rear axle to the mass centre (m)
    (X,Y) is the cartesian position of the vehicle
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 8   
        self.m = 3   
        self.p = 8
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Full Dynamic Bicyle Model with slip'
        self.state_label = ['v_x','v_y','dtheta','theta','X','Y','omega_f','omega_r']
        self.input_label = ['delta', 'T_f','T_r']
        self.output_label = ['v_x','v_y','dtheta','theta','X','Y','omega_f','omega_r']
        
        # Units
        self.state_units = ['[m/s]','[m/s]','[rad/s]','[rad]','[m]','[m]','[rad/s]','[rad/s]']
        self.input_units = ['[rad]', '[Nm]','[Nm]']
        self.output_units = ['[m/s]','[m/s]','[rad/s]','[rad]','[m]','[m]','[rad/s]','[rad/s]']

        
        # Model param (used for animation purposes only change the values)
        self.width  = 2.00
        self.a      = 2.00
        self.b      = 3.00
        self.lenght = self.a+self.b    
        self.lenght_tire = 0.40
        self.width_tire = 0.15
        
        self.g = 9.81
        self.mass = 1000
        self.Iz = 1.00/12.00*self.mass*((self.a+self.b)**2+self.width**2) 
        self.Iprop_f = 20
        self.Iprop_r = 20
        self.r = 0.3
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = 5
        
    #############################
    def combinedSlipTireModel(self,x,u):
        
        #Tire-road friction coefficient
        self.mu = 0.1
        F_nf = self.mass*self.g*self.b/(self.b+self.a)
        F_nr = self.mass*self.g*self.a/(self.b+self.a)
        #Compute the max forces available
        max_F_f = F_nf*self.mu
        max_F_r = F_nr*self.mu
        #Compute the lateral "slip-slope"
        self.max_alpha_stat = 0.12
        slip_ratio_f = max_F_f/(self.max_alpha_stat)
        slip_ratio_r = max_F_r/(self.max_alpha_stat)
        # Simulation of road-tire behavior according to slip-ratios
        self.max_slip_stat = 0.03
        slip_slope_f = max_F_f/self.max_slip_stat
        slip_slope_r = max_F_r/self.max_slip_stat
        # Compute slip ratios and longitudinal forces for both tires (depend on Vx (x[0]), omega_f (x[6]), omega_r (x[7]))
        if(x[0] == 0 or x[6] == 0):
            long_slip_f = 0
            F_xf = 0
        elif(u[1] == 0):
            long_slip_f = 0
            F_xf = 0
        else:
            long_slip_f = (x[6]*self.r-x[0])/np.maximum(np.absolute(x[0]),np.absolute(x[6]*self.r))  
            if (long_slip_f<-0.03):
                F_xf = -max_F_f
            elif(long_slip_f>0.03):
                F_xf = max_F_f
            else:
                F_xf = slip_slope_f*long_slip_f
        if(x[0] == 0 and x[7] == 0):
            F_xr = 0
            long_slip_r = 0
        elif(u[2] == 0):
            long_slip_r = 0
            F_xr = 0
        else:
            long_slip_r = (x[7]*self.r-x[0])/max(x[0],x[7]*self.r)
            if (long_slip_r<-0.03):
                F_xr = -max_F_r
            elif(long_slip_r>0.03):
                F_xr = max_F_r
            else:
                F_xr = slip_slope_r*long_slip_r
        # Compute slip angles and lateral forces for both tires (depends on Vx (x[0]),Vy (x[1]) and delta (u[0]))
        if x[0] == 0:
            slip_f = 0
            slip_r = 0
        elif(x[0]>0):
            slip_f = np.arctan((x[1]+self.a*x[2])/x[0])-u[0]
            slip_r = np.arctan((x[1]-self.b*x[2])/x[0])
        else:
            slip_f = -(np.arctan((x[1]+self.a*x[2])/x[0])-u[0])
            slip_r = -np.arctan((x[1]-self.b*x[2])/x[0])
        if (slip_f<-0.12):
            F_yf = max_F_f
        elif (slip_f > 0.12):
            F_yf = -max_F_f
        else:
            F_yf = -slip_ratio_f*slip_f
            
        if (slip_r<-0.1):
            F_yr = max_F_r
        elif (slip_r > 0.1):
            F_yr = -max_F_r
        else:
            F_yr = -slip_ratio_r*slip_r
        
        return F_yf,F_yr,F_xf,F_xr  
    #############################
        
    def f(self, x = np.zeros(8) , u = np.zeros(3) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        dx = np.zeros(self.n) # State derivative vector
        F_yf,F_yr,F_xf,F_xr = self.combinedSlipTireModel(x,u)
        
        dx[0] = (F_xf*np.cos(u[0])-F_yf*np.sin(u[0])+F_xr)/self.mass-x[1]*x[2]
        dx[1] = (F_yf*np.cos(u[0])+F_xf*np.sin(u[0])+F_yr)/self.mass-x[0]*x[2]
        dx[2] = (self.a*(F_yf*np.cos(u[0])+F_xf*np.sin(u[0]))-self.b*F_yr)/self.Iz
        dx[3] = x[2]
        dx[4] = x[0]*np.cos(x[3])-x[1]*np.sin(x[3])
        dx[5] = x[0]*np.sin(x[3])+x[1]*np.cos(x[3])
        dx[6] = (u[1]-self.r*F_xf)/self.Iprop_f
        dx[7] = (u[2]-self.r*F_xr)/self.Iprop_r
        
        return dx
    
    ############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q =  np.array( [ x[4] ,  x[5] , x[3] , u[0] ] )
        
        return q

##############################################################################
        
class FullDynamicBicycleModelwithVoltInput( LateralDynamicBicycleModelwithSpeedInput ):
    
    """ 
    Equations of Motion
    -------------------------
    STATES_DERIVATIVE
    dv_x    = (F_xf*cos(delta)-F_yf*sin(delta)+F_xr)/m - v_y*dtheta   "longitudinal force sum (F_x = m*dv_x) gives longitudinal acceleration dv_x"
    dv_y    = (F_yf*cos(delta)+F_xf*sin(delta)+F_yr)/m - v_x*dtheta "lateral force sum (F_y = m*dv_y) gives lateral acceleration dv_y"
    ddtheta = (a*(F_yf*cos(delta)+F_xf*sin(delta))-b*F_yr)/Iz       "Torque sum at mass center (T = I*ddtheta) gives the angular acceleration of the vehicle ddtheta"
    dtheta  = dtheta                                                "dtheta is already a state which is the yaw rate"
    dX      = v_x*cos(theta)-v_y*sin(theta)                         "To obtain cartesian position"
    dY      = v_x*sin(theta)+v_y*cos(theta)
    domega_r = ((V-omega_r*gR*k)*gR*k/R - F_xr*r - B(omega))/J      "Propulsion model"
    
    ***** DEPENDS DU MODELE DE TIRE COMPLET EX: S=omega/v_x et F_x,F_y = f(F_z,alpha,S et mu) *****
    
    INPUTS 
    delta is the steering angle (rad)
    propulsion voltage (V)
    
    
    Where 
    F_yf is the lateral force applied perpendicularly with the front wheel (N)
    F_yr is the lateral force applied perpendicularly with the rear wheel (N)
    v_y  is the lateral velocity of the mass center (m/s)
    v_x  is the longitudinal velocity of the mass center (m/s)
    delta is the steering angle (rad)
    theta is the yaw angle of the vehicle (rad)
    m     is the mass of the vehicle (kg)
    Iz    is the inertia for a rotation along the z-axis (kg*m^2)
    a     is the distance from the front axle to the mass centre (m)
    b     is the distance from the rear axle to the mass centre (m)
    (X,Y) is the cartesian position of the vehicle
    r     is the wheels radius (m)
    gR    is the gear ratio
    R     is the motor's resistance (ohm)
    k     is the motor constant (Nm/A)
    J     is the propulsion's inertia
    B(omega) is the viscuous friction function
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 7   
        self.m = 2   
        self.p = 7
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Full Dynamic Bicyle Model with slip'
        self.state_label = ['v_x','v_y','dtheta','theta','X','Y','omega_r']
        self.input_label = ['delta', 'V']
        self.output_label = ['v_x','v_y','dtheta','theta','X','Y','omega_r']
        
        # Units
        self.state_units = ['[m/s]','[m/s]','[rad/s]','[rad]','[m]','[m]','[rad/s]']
        self.input_units = ['[rad]', '[Nm]','[Nm]']
        self.output_units = ['[m/s]','[m/s]','[rad/s]','[rad]','[m]','[m]','[rad/s]']

        
        # Model param (used for animation purposes only change the values)
        self.width  = 0.2
        self.a      = 0.204
        self.b      = 0.136
        self.lenght = self.a+self.b    
        self.lenght_tire = 0.11
        self.width_tire = 0.04
        
        self.g = 9.81
        self.mass = 4.1
        self.Iz = 1.00/12.00*self.mass*((self.a+self.b)**2+self.width**2) 
        self.r = 0.055
        
        # Propulsion param
        self.R = 0.63
        self.k = 0.002791
        self.J = 0.120881
        self.gR = 15.3
        
        # Graphic output parameters 
        self.dynamic_domain  = True
        self.dynamic_range   = 5
        
    #############################
    def combinedSlipTireModel(self,x,u):
        
        #Tire-road friction coefficient
        self.mu = 0.45
        F_nf = self.mass*self.g*self.b/(self.b+self.a)
        F_nr = self.mass*self.g*self.a/(self.b+self.a)
        #Compute the max forces available
        max_F_f = F_nf*self.mu
        max_F_r = F_nr*self.mu
        #Compute the lateral "slip-slope"
        self.max_alpha_lin = 0.12
        slip_ratio_f = -max_F_f/(self.max_alpha_lin)
        slip_ratio_r = -max_F_r/(self.max_alpha_lin)
        # Simulation of road-tire behavior according to slip-ratios
        self.max_slip_lin = 0.07
        slip_slope_r = max_F_r/self.max_slip_lin
        # Compute slip ratios and longitudinal forces for both tires (depend on Vx (x[0]), omega_f (x[6]), omega_r (x[7]))
        F_xf = 0
        if(abs(x[0]) < 0.1 and abs(x[6]) < 0.1):
            F_xr = 0
            long_slip_r = 0
        else:
            long_slip_r = (x[6]*self.r-x[0])/max(abs(x[0]),abs(x[6]*self.r))
            if (long_slip_r<-self.max_slip_lin):
                F_xr = -max_F_r
            elif(long_slip_r>self.max_slip_lin):
                F_xr = max_F_r
            else:
                F_xr = slip_slope_r*long_slip_r

        # Compute slip angles and lateral forces for both tires (depends on Vx (x[0]),Vy (x[1]) and delta (u[0]))
        if abs(x[0]) < 0.1:
            slip_f = 0
            slip_r = 0
        else:
            slip_f = np.arctan((x[1]+self.a*x[2])/x[0])-u[0]
            slip_r = np.arctan((x[1]-self.b*x[2])/x[0])

        if (slip_f<-self.max_alpha_lin):
            F_yf = max_F_f
        elif (slip_f > self.max_alpha_lin):
            F_yf = -max_F_f
        else:
            F_yf = slip_ratio_f*slip_f
            
        if (slip_r<-self.max_alpha_lin):
            F_yr = max_F_r
        elif (slip_r > self.max_alpha_lin):
            F_yr = -max_F_r
        else:
            F_yr = slip_ratio_r*slip_r
            
            
        return F_yf,F_yr,F_xf,F_xr  
    #############################
        
    def f(self, x = np.zeros(7) , u = np.zeros(2) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        dx = np.zeros(self.n) # State derivative vector
        F_yf,F_yr,F_xf,F_xr = self.combinedSlipTireModel(x,u)
        
        dx[0] = (F_xf*np.cos(u[0])-F_yf*np.sin(u[0])+F_xr)/self.mass + x[1]*x[2]
        dx[1] = (F_yf*np.cos(u[0])+F_xf*np.sin(u[0])+F_yr)/self.mass - x[0]*x[2]
        dx[2] = (self.a*(F_yf*np.cos(u[0])+F_xf*np.sin(u[0]))-self.b*F_yr)/self.Iz
        dx[3] = x[2]
        dx[4] = x[0]*np.cos(x[3])-x[1]*np.sin(x[3])
        dx[5] = x[0]*np.sin(x[3])+x[1]*np.cos(x[3])
        dx[6] = ((u[1]-x[6]*self.gR*self.k)*self.gR*self.k/self.R-F_xr*self.r)/self.J
        
        return dx
    
    ############################
    def xut2q( self, x , u , t ):
        """ compute config q """
        
        q =  np.array( [ x[4] ,  x[5] , x[3] , u[0] ] )
        
        return q  
        
    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    sys = LateralDynamicBicycleModelwithSpeedInput()
    sys.ubar = np.array([1.1,10])
    
    sys.plot_trajectory()
    sys.animate_simulation()
    
    sys      = LateralDynamicBicycleModelwithForceInputs()
    sys.ubar = np.array([0.8,000,200])
    
    sys.plot_trajectory()
    sys.animate_simulation()
    
    sys      = FullDynamicBicycleModelwithTorquesInputs()
    sys.ubar = np.array([0.2,0,1000])
    
    sys.plot_trajectory()
    sys.animate_simulation()
    
        


# -*- coding: utf-8 -*-

###############################################################################
import numpy as np
###############################################################################
from pyro.control.controller import StaticController
###############################################################################

class stable_baseline3_controller( StaticController ) :
    """ 
    Wrap a stable baseline 3 model to use it as a pyro controller
    """

    def __init__(self, model):

        self.model = model

        # Dimensions
        self.k = model.observation_space.shape[0]
        self.m = model.action_space.shape[0]
        self.p = model.observation_space.shape[0]
        
        StaticController.__init__( self, self.k, self.m, self.p)

        self.name = "Stable Baseline 3 Controller"


    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c( y, r, t)
        
        INPUTS
        y  : sensor signal vector          p x 1
        r  : reference signal vector       k x 1
        t  : time                          1 x 1
        
        OUTPUTS
        u  : control inputs vector         m x 1
        
        """
        
        u, _x = self.model.predict( y , deterministic = True )
        
        return u

        


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.control  import nonlinear
from pyro.analysis import simulation
###############################################################################

sys  = pendulum.DoublePendulum()


# ref traj
#traj = simulation.Trajectory.load( 'double_pendulum_rrt.npy' )
#traj = simulation.Trajectory.load( 'double_pendulum_directcollocation.npy' )
traj = simulation.Trajectory.load('double_pendulum_directcollocation_hires.npy')

# Controller
ctl  = nonlinear.SlidingModeController( sys , traj )
#ctl  = nonlinear.SlidingModeController( sys )

# goal
ctl.rbar = np.array([0,0])
ctl.lam  = 1
ctl.gain = 2

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0  = np.array([-3.14,0,0,0])


# Solver param
tf = 10
n  = 10001
solver = 'euler' # necessary for sliding mode

cl_sys.compute_trajectory(tf,n,solver)
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory(0, 2)
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov  9 10:46:14 2018

@author: alxgr
"""

###############################################################################
import numpy as np
from scipy.interpolate import interp1d
###############################################################################
from pyro.control import controller
from pyro.dynamic import mechanical
###############################################################################




###############################################################################
# Computed Torque
###############################################################################
        
class ComputedTorqueController( controller.StaticController ) :
    """ 
    Inverse dynamic controller for mechanical system

    """    
    
    ############################
    def __init__(self, model = mechanical.MechanicalSystem() , traj = None ):
        """
        
        ---------------------------------------
        r  : reference signal vector  k x 1
        y  : sensor signal vector     p x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        ---------------------------------------
        u = c( y , r , t )
        
        """
        
        self.model = model
        
        # Dimensions
        self.k = model.dof   
        self.m = model.m
        self.p = model.p
        
        controller.StaticController.__init__(self, self.k, self.m, self.p)
        
        # Label
        self.name = 'Computed Torque Controller'
        
        # Params
        self.w0   = 1
        self.zeta = 0.7 
        
        # Mode
        if traj == None:
            self.c = self.c_fixed_goal
        else:
            self.load_trajectory( traj )
            self.mode = 'interpol'
            self.c = self.c_trajectory_following
        
    
    #############################
    def c_fixed_goal( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
                
        x   = y 
        q_d = r
        
        u = self.fixed_goal_ctl( x , q_d , t )
        
        return u
    
        
        
    ############################
    def fixed_goal_ctl( self , x , q_d , t = 0 ):
        """ 
        
        Given desired fixed goal state and actual state, compute torques
        
        """
        [ q , dq ]     = self.model.x2q( x )  
        
        ddq_d          = np.zeros( self.model.dof )
        dq_d           = np.zeros( self.model.dof )

        ddq_r          = self.compute_ddq_r( ddq_d , dq_d , q_d , dq , q )
        
        u              = self.model.actuator_forces( q , dq , ddq_r )
        
        return u
        
        
    ############################
    def compute_ddq_r( self , ddq_d , dq_d , q_d , dq , q ):
        """ 
        
        Given desired trajectory and actual state, compute ddq_r
        
        """
        
        q_e   = q  -  q_d
        dq_e  = dq - dq_d
        
        ddq_r = ddq_d - 2 * self.zeta * self.w0 * dq_e - self.w0 ** 2 * q_e
        
        return ddq_r
    
        
    ############################
    def load_trajectory( self , traj  ):
        """ 
        
        Load Open-Loop trajectory solution to use as reference trajectory
        
        """
        
        self.trajectory = traj
        
        q   = traj.x[ :,    0           :     self.model.dof ]
        dq  = traj.x[ :, self.model.dof : 2 * self.model.dof ]
        ddq = traj.dx[:, self.model.dof : 2 * self.model.dof ]
        t   = traj.t
        
        # Create interpol functions
        self.q   = interp1d(t,q.T)
        self.dq  = interp1d(t,dq.T)
        self.ddq = interp1d(t,ddq.T)
        
        
    ############################
    def get_traj( self , t  ):
        """ 
        
        Find closest point on the trajectory
        
        """
        
        if t < self.trajectory.time_final :

            # Load trajectory
            q     = self.q(   t )
            dq    = self.dq(  t )
            ddq   = self.ddq( t )          

        else:
            
            q     = self.rbar
            dq    = np.zeros( self.model.dof )
            ddq   = np.zeros( self.model.dof )
        
        return ddq , dq , q
    
    
    ############################
    def traj_following_ctl( self , x , t = 0 ):
        """ 
        
        Given desired loaded trajectory and actual state, compute torques
        
        """
        
        [ q , dq ]         = self.model.x2q( x ) 
        
        ddq_d , dq_d , q_d = self.get_traj( t )

        ddq_r              = self.compute_ddq_r( ddq_d , dq_d , q_d , dq , q )
        
        u                  = self.model.actuator_forces( q , dq , ddq_r )
        
        return u
        
        
    #############################
    def c_trajectory_following( self , y , r , t ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        x = y 
        
        u = self.traj_following_ctl( x , t )
        
        
        return u
    


##############################################################################
        
class SlidingModeController( ComputedTorqueController ):
    """ 
    Sliding Mode Controller for fully actuated mechanical systems
    """
    
    
    ############################
    def __init__( self , model , traj = None ):
        """ """
        
        ComputedTorqueController.__init__( self, model , traj )
        
        # Params
        self.lam  = 1   # Sliding surface slope
        self.gain = 1   # Discontinuous gain
        self.nab  = 0.1 # Min convergence rate
        
        
    ############################
    def compute_sliding_variables( self , ddq_d , dq_d , q_d , dq , q ):
        """ 
        
        Given desired trajectory and actual state
        
        """        
        q_e   = q  -  q_d
        dq_e  = dq - dq_d
        
        s      = dq_e  + self.lam * q_e
        dq_r   = dq_d  - self.lam * q_e
        ddq_r  = ddq_d - self.lam * dq_e
        
        return [ s , dq_r , ddq_r ]
        
        
    ############################
    def K( self , q , t ):
        """ Discontinuous gain matrix """
        
        dist_max = np.diag( np.ones( self.model.dof ) ) * self.gain
        conv_min = np.diag( np.ones( self.model.dof ) ) * self.nab
        
        K = dist_max + np.dot( self.model.H( q ) , conv_min ) 
        
        return K
        
        
    ############################
    def sliding_torque( self , ddq_r , s , dq , q , t ):
        """ 
        
        Given actual state, compute torque necessarly to guarantee convergence
        
        """
                
        u_computed      = self.model.actuator_forces( q , dq , ddq_r )
        
        u_discontinuous = np.dot( self.K( q , t ) ,  np.sign( s ) )
        
        u_tot = u_computed - u_discontinuous
        
        return u_tot
        
        
    ############################
    def traj_following_ctl( self , x , t = 0 ):
        """ 
        
        Given desired loaded trajectory and actual state, compute torques
        
        """
        
        [ q , dq ]            = self.model.x2q( x ) 
        
        ddq_d , dq_d , q_d    = self.get_traj( t )

        [ s , dq_r , ddq_r ]  = self.compute_sliding_variables( ddq_d , dq_d , 
                                                                q_d , dq , q )
        
        u                     = self.sliding_torque( ddq_r , s , dq , q , t )
        
        return u
        
        
    ############################
    def fixed_goal_ctl( self , x , q_d , t = 0 ):
        """ 
        
        Given desired fixed goal state and actual state, compute torques
        
        """
        
        [ q , dq ]     = self.model.x2q( x ) 
        
        ddq_d          =   np.zeros( self.model.dof )
        dq_d           =   np.zeros( self.model.dof )

        [ s , dq_r , ddq_r ]  = self.compute_sliding_variables( ddq_d , dq_d , 
                                                                q_d , dq , q )
        
        u                     = self.sliding_torque( ddq_r , s , dq , q , t )
        
        return u

    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    from pyro.dynamic import pendulum


    sys  = pendulum.DoublePendulum()
    ctl  = ComputedTorqueController( sys )
    
    # New cl-dynamic
    cl_sys = ctl + sys
    
    cl_sys.x0 = np.array([2,1,0,0])
    cl_sys.compute_trajectory()
    cl_sys.plot_phase_plane_trajectory(0,2)
    cl_sys.animate_simulation()
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import drone
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = drone.ConstantSpeedHelicopterTunnel()

sys.obstacles = [
                [ (2,2),(4,4)],
                [ (8,5),(10,10)],
                [ (14,0),(16,4)]
                ]

sys.mass = 0.1
sys.vx   = 5.0
sys.width =1.0

# Set domain
sys.x_ub = np.array([+60, 10, +20])
sys.x_lb = np.array([-60, 0, +0])

sys.u_ub = np.array([+20])
sys.u_lb = np.array([-20])

"""
sys.show([0,0])dp.

sys.ubar[0] = 10
sys.animate_simulation()
"""


# Discrete world
grid_sys = discretizer.GridDynamicSystem(sys, (51, 51, 51), [11], 0.05)

# Cost Function
qcf = costfunction.QuadraticCostFunctionWithDomainCheck.from_sys( sys )

qcf.xbar = np.array([ 0.0 , 2.0, 20 ]) # target
qcf.INF  = 100000
qcf.EPS  = 0.2

qcf.Q[0,0] = 2.0
qcf.Q[1,1] = 200.0
qcf.Q[2,2] = 0.0

qcf.R[0,0] = 5.0

qcf.S[0,0] = 20.0
qcf.S[1,1] = 50.0
qcf.S[2,2] = 0.0

# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.plot_cost2go(qcf.INF,2,1)

dp.alpha = 0.999
dp.solve_bellman_equation()
    

dp.plot_cost2go(qcf.INF,2,1)


ctl = dp.get_lookup_table_controller()

cl_sys = ctl + sys

# Simulation and animation
cl_sys.x0   = np.array([0,8,-10])
cl_sys.compute_trajectory( 6, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:01:07 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
import advanced_vehicles
import test_vehicle_controllers
###############################################################################

# "Fake" controller - Varying inputs (delta, T_f, T_r) throughout time (change in linear.py)
ctl = test_vehicle_controllers.fullDynTorqueInputs()
# Vehicule dynamical system
sys = advanced_vehicles.FullDynamicBicycleModelwithTorquesInputs()
# Add the inputs to the dynamical system
cl_sys = ctl + sys
# Plot open-loop behavior (ex: np.array[intial_conditions], time_of_simulation)
cl_sys.x0 =  np.array([5,0,0,0,0,0,0,5/0.3])
cl_sys.compute_trajectory( 50 )
cl_sys.plot_trajectory('xu')

# Rebuild x,u and t from simulation
x = cl_sys.traj.x
u = cl_sys.traj.u 
t = cl_sys.traj.t    

# Compute tire forces and slip angles for graphical purpose  
F_nf = sys.mass*sys.g*sys.b/(sys.b+sys.a)
F_nr = sys.mass*sys.g*sys.a/(sys.b+sys.a)
mu = sys.mu
max_F_f = F_nf*mu
max_F_r = F_nr*mu
max_alpha_stat = sys.max_alpha_stat
slip_ratio_f = max_F_f/(max_alpha_stat) #0.12 is the max slipping angle in rad before kinetic slip
slip_ratio_r = max_F_r/(max_alpha_stat) 
# Simulation of road-tire behavior according to slip-ratios
max_slip_stat = sys.max_slip_stat
slip_slope_f = max_F_f/max_slip_stat
slip_slope_r = max_F_r/max_slip_stat
# Init. vectors for display purposes
F_yf = np.zeros(len(t))
F_yr = np.zeros(len(t))
slip_f = np.zeros(len(t))
slip_r = np.zeros(len(t))
maxF_f = np.zeros(len(t))
maxF_r = np.zeros(len(t))
long_slip_f = np.zeros(len(t))
F_xf = np.zeros(len(t))
F_xr = np.zeros(len(t))
long_slip_r = np.zeros(len(t))
# Iterate through time to compute slip angles, slip ratios, lateral forces and longitudinal forces
for i in range(len(t)-1):
    # Compute slip ratios and longitudinal forces for both tires (depend on Vx (x[0]), omega_f (x[6]), omega_r (x[7]))
    if(x[i,0] == 0 or x[i,6] == 0):
        long_slip_f[i] = 0
        F_xf[i] = 0
    elif(u[i,1] == 0):
        long_slip_f[i] = 0
        F_xf[i] = 0
    else:
        long_slip_f[i] = (x[i,6]*sys.r-x[i,0])/np.maximum(np.absolute(x[i,0]),np.absolute(x[i,6]*sys.r))  
        if (long_slip_f[i]<-0.03):
            F_xf[i] = -max_F_f
        elif(long_slip_f[i]>0.03):
            F_xf[i] = max_F_f
        else:
            F_xf[i] = slip_slope_f*long_slip_f[i]
    if(x[i,0] == 0 and x[i,7] == 0):
        F_xr[i] = 0
        long_slip_r[i] = 0
    elif(u[i,2] == 0):
        long_slip_r[i] = 0
        F_xr[i] = 0
    else:
        long_slip_r[i] = (x[i,7]*sys.r-x[i,0])/max(x[i,0],x[i,7]*sys.r)
        if (long_slip_r[i]<-0.03):
            F_xr[i] = -max_F_r
        elif(long_slip_r[i]>0.03):
            F_xr[i] = max_F_r
        else:
            F_xr[i] = slip_slope_r*long_slip_r[i]
    maxF_f[i] = max_F_f
    maxF_r[i] = max_F_r

# Plot forces
figsize   = (7, 4)
dpi       = 100
plt.figure(2, figsize=figsize, dpi=dpi)
plt.title('Longitudinal forces for the longitudinal dynamic\n model with steering angle and torque as inputs', fontsize=20)
plt.plot(t[:-1], F_xf[:-1], label = 'F_xf')
plt.plot(t[:-1], F_xr[:-1], label = 'F_xr')
plt.plot(t[:-1], maxF_f[:-1],'--', label = 'Max front force')
plt.plot(t[:-1], maxF_r[:-1],'--', label = 'Max rear force')
plt.plot(t[:-1], -maxF_r[:-1],'--', label = 'Max rear force')
plt.legend(fontsize = '15')
plt.xlabel('Temps (s)')
plt.ylabel('Force (N)')
plt.show()


# Plot trajectory of the vehicle's CG
slip = 1
t_start_slip = 'No slip' #No slip
plt.figure(3,figsize=figsize, dpi=dpi)
for i in range(len(t)-1):
    if (F_xf[i]==max_F_f or F_xr[i]==max_F_r):
        if (slip == 1):            
            t_start_slip = i
            slip = 0
        else:
            t_stop_slip = i
    else:
        pass
if (t_start_slip == 'No slip'):
    plt.plot(x[:,4],x[:,5],'-b', label = 'No slip')  
else:
    plt.plot((x[t_start_slip:t_stop_slip,4]),(x[t_start_slip:t_stop_slip,5]),'-r',label='Slip')
    plt.plot((x[0:t_start_slip,4]),(x[0:t_start_slip,5]),'-b', label='No slip')
    plt.plot((x[t_stop_slip:len(t)-1,4]),(x[t_stop_slip:len(t)-1,5]),'-b')    
plt.legend(fontsize ='15')
plt.show()

# Plot slip_ratios according to time
plt.figure(4, figsize=figsize, dpi=dpi)
plt.title('Slip ratios', fontsize=20)
plt.plot(t[:-1], long_slip_f[:-1], label = 'Sx_f')
plt.plot(t[:-1], long_slip_r[:-1], label = 'Sx_r')
plt.legend(fontsize = '15')
plt.xlabel('Temps (s)')
plt.ylabel('Slip ratio')
plt.show()

# Animate the simulation
cl_sys.plot_trajectory('x')
cl_sys.plot_trajectory('u')
cl_sys.animate_simulation()




# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

torque_controlled_robot.ubar = np.array([2,0.8]) # constant inputs
torque_controlled_robot.x0   = np.array([0,1,0,0])

torque_controlled_robot.plot_trajectory('xu')
anim1 = torque_controlled_robot.animate_simulation()


speed_controlled_robot      = manipulator.SpeedControlledManipulator.from_manipulator( 
                                             torque_controlled_robot )
speed_controlled_robot.ubar = np.array([2,0.8]) # constant inputs
speed_controlled_robot.x0   = np.array([0,1])

speed_controlled_robot.plot_trajectory('xu')
anim2 = speed_controlled_robot.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 24 21:15:49 2020

@author: alex
"""

import numpy as np
from pyro.control import controller


##############################################################################
class SinglePendulumAdaptativeController( controller.DynamicController ):
    """ 
    Adaptative Controller for fully actuated mechanical systems (single pendulum)
    """
    ############################
    def __init__( self , model , traj = None ):
        """ """ 
        self.name = 'Adaptive controller'

        # Params
        self.P    = np.eye(2)
        self.K    = 1
        self.lam  = 1   # gain
        
        self.model=model
        
        k = model.dof   
        m = model.m
        p = model.p
        
        l = 2 # number of states in the controller
        
        super().__init__(k, l, m, p)
        
        # Init internal states
        self.z0 = np.array([0.0,0.0])
        
        
        self.internal_state_label = []
        self.internal_state_units = []
        
        for i in range(l):
            self.internal_state_label.append('a' +str(i))
            self.internal_state_units.append('')
            
        
    ############################
    def adaptative_variables( self , ddq_d , dq_d , q_d , dq , q ):
        """ 
        
        Compute intermediate variables
        
        """        
        q_e   = q_d  - q
        dq_e  = dq_d - dq
        
        s      = dq_e  + self.lam * q_e
        dq_r   = dq_d  + self.lam * q_e
        ddq_r  = ddq_d + self.lam * dq_e
        
        Y_r = np.zeros(2)

        Y_r[0] = ddq_r
        Y_r[1] = np.sin(q)
        
        return [ s , dq_r , ddq_r , Y_r ]
    
                        
    ############################
    def b(self, a , x , r , t):
        """
        adaptation law

        """
        
        [ q , dq ]                  = self.model.x2q( x ) 
        
        [ ddq_d , dq_d , q_d ]      = self.get_traj( t , r )
        
        [ s , dq_r , ddq_r , Y_r ]  = self.adaptative_variables( ddq_d , dq_d , q_d , dq , q )

        da   = self.P @ Y_r * s
        
        return da
    
    
    ############################
    def c(self , a , x , r , t = 0):
        """ 
        
        control law
        
        """
        [ q , dq ]                  = self.model.x2q( x ) 
        
        [ ddq_d , dq_d , q_d ]      = self.get_traj( t , r )
        
        [ s , dq_r , ddq_r , Y_r ]  = self.adaptative_variables( ddq_d , dq_d , q_d , dq , q )
                
        u   = Y_r @ a + self.K * s
        
        return u
    
    
    ############################
    def get_traj(self, t , r ):
        """  """
        
        ddq_d          =   np.zeros( self.model.dof )
        dq_d           =   np.zeros( self.model.dof )
        q_d            =   r
        
        return [ ddq_d , dq_d , q_d ] 
    



##############################################################################
        
class DoublePendulumAdaptativeController(  controller.DynamicController ):
    """ 
    
    """
    
    ############################
    def __init__( self , model , traj = None ):
        """ """
        
        self.name = 'Adaptive controller'

        self.A = np.zeros(5)
        self.guess = np.zeros(5)
        self.T=np.eye(5)
        self.Kd = np.eye(2)
        self.lam  = 1   # Sliding surface slope
        self.nab  = 0.1 # Min convergence rate
        
        self.model=model
        
        k = model.dof   
        m = model.m
        p = model.p
        l = 5
        
        super().__init__(k, l, m, p)
        
        
        
        # Init states
        self.z0 = np.zeros(5)
        
        self.internal_state_label = []
        self.internal_state_units = []
        
        for i in range(l):
            self.internal_state_label.append('a' +str(i))
            self.internal_state_units.append('')
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )
        c12 = np.cos( q[0] + q[1] )
        s12 = np.sin( q[0] + q[1] )
        
        return [c1,s1,c2,s2,c12,s12]

        
    ############################
    def adaptative_variables( self , ddq_d , dq_d , q_d , dq , q ):
        """ 
        
        Given desired trajectory and actual state
        
        """        
        q_e   = q  -  q_d
        dq_e  = dq - dq_d
        
        s      = dq_e  + self.lam * q_e
        dq_r   = dq_d  - self.lam * q_e
        ddq_r  = ddq_d - self.lam * dq_e
        
        return [ s , dq_r , ddq_r ]
        
        
    ############################
    def adaptative_torque( self , Y , s , q , t ):
        """ 
        
        Given actual state, compute torque necessarly to guarantee convergence
        
        """
                
        u_computed      = np.dot( Y , self.A  )
        
        u_discontinuous = np.dot(self.Kd,s)
        
        u_tot = u_computed - u_discontinuous
        
        return u_tot
    
                        
    ############################
    def b(self, z, x, q_d, t):
        
        [ q , dq ]     = self.model.x2q( x ) 
        
        ddq_d          =   np.zeros( self.model.dof )
        dq_d           =   np.zeros( self.model.dof )
        
        [ s , dq_r , ddq_r ]  = self.adaptative_variables( ddq_d , dq_d , 
                                                           q_d , dq , q )
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        Y = np.zeros((2,5))
        dz = np.zeros(5)
                
        Y[0,0]=ddq_r[0]*c2
        Y[0,1]=ddq_r[1]*c2
        Y[0,2]=s2*dq[1]*dq_r[0]
        Y[0,3]=s2*(dq[0]+dq[1])*dq_r[1]
        Y[0,4]=s1+s12
        Y[1,0]=ddq_r[0]*c2
        Y[1,1]=ddq_r[1]
        Y[1,2]=s2*dq[0]*dq_r[0]
        Y[1,3]=0
        Y[1,4]=s12
        
        b = np.dot(Y.T,s)
        dz=-1*np.dot( self.T , b )
        
        return dz
    
        
    ############################
    def c( self , z , x , q_d , t = 0 ):
        """ 
        
        Given desired fixed goal state and actual state, compute torques
        
        """
        
        [ q , dq ]     = self.model.x2q( x ) 
        
        ddq_d          =   np.zeros( self.model.dof )
        dq_d           =   np.zeros( self.model.dof )
        
        [ s , dq_r , ddq_r ]  = self.adaptative_variables( ddq_d , dq_d , 
                                                           q_d , dq , q )
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        Y = np.zeros((2,5))
                
        Y[0,0]=ddq_r[0]*c2
        Y[0,1]=ddq_r[1]*c2
        Y[0,2]=s2*dq[1]*dq_r[0]
        Y[0,3]=s2*(dq[0]+dq[1])*dq_r[1]
        Y[0,4]=s1+s12
        Y[1,0]=ddq_r[0]*c2
        Y[1,1]=ddq_r[1]
        Y[1,2]=s2*dq[0]*dq_r[0]
        Y[1,3]=0
        Y[1,4]=s12
        
        self.A= self.guess + self.get_z_integral(z)
                
        u                     = self.adaptative_torque(  Y , s  , q , t )
        
        return u
    
    
    ############################
    def get_z_integral(self, z):
        """ get intergral error internal states """
        
        return z[:self.l]
    
        
##############################################################################


###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from asimov import Asimov
from pyro.control import nonlinear
###############################################################################

asimov = Asimov()  # Asimov
asimov.x0 = np.array([-np.pi/4, -3*np.pi/4, np.pi/2, 0, 0, 0])  # Position initiale

ctl = nonlinear.ComputedTorqueController(asimov)  # Déclaration du controlleur
ctl.rbar = np.array([0.5, 0.25, 0])  # Cible
ctl.w0 = 2
ctl.zeta = 1

closed_loop_robot = ctl + asimov  # Système boucle fermé

closed_loop_robot.plot_trajectory()  # Calcul de la trajectoire
closed_loop_robot.animate_simulation( is_3d=True )



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 24 13:07:39 2018

@author: nvidia
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import mechanical
###############################################################################


###############################################################################

class SinglePendulum( mechanical.MechanicalSystem ):
    """Pendulum with a point mass and inertial rod.

    Attributes
    ----------
    l1 : float
        Length of pendulum rod. Only used for display.
    lc1 : float
        Distance of point mass to pivot.
    m1 : float
        Mass value of point mass.
    I1 : float
        Moment of inertia of pendulum rod (without mass) about the pendulum pivot. The
        total inertia of the pendulum is calculated as
        ``I_pendulum = I1 + (m1 * lc1**2)``.
    gravity : float
        Constant of gravitational acceleration
    d1: float
        Damping force factor
    """

    ############################
    def __init__(self):
        """ """
               
        # initialize standard params
        mechanical.MechanicalSystem.__init__(self, 1)
        
        # Name
        self.name = 'Single Pendulum'
        
        # params
        self.setparams()
        
            
    #############################
    def setparams(self):
        """ Set model parameters here """
        
        # kinematic
        self.l1  = 2.0 
        self.lc1 = 1
        
        # dynamic
        self.m1       = 1
        self.I1       = 1
        self.gravity  = 9.81
        self.d1       = 0
        
        # graphic
        self.l_domain = 5.0
        
        
    ##############################
    def trig(self, q ):
        """ Compute cos and sin """
        
        c1  = np.cos( q )
        s1  = np.sin( q )

        return [c1,s1]
    
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        H = np.zeros((self.dof,self.dof))
        
        H[0,0] = self.m1 * self.lc1**2 + self.I1
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        C = np.zeros((self.dof,self.dof))

        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.diag( np.ones( self.dof ) ) #  identity matrix
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros( self.dof ) 
        
        [c1,s1] = self.trig( q )
        
        g[0] = self.m1 * self.gravity * self.lc1 * s1

        return g
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d    = np.zeros( self.dof ) 
        
        d[0] = self.d1 * dq[0]
        
        return d
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = self.l_domain
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        # ground line
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([-10,0,0])
        pts[1,:] = np.array([+10,0,0])
        
        lines_pts.append( pts )
        lines_style.append( '--')
        lines_color.append( 'k' )
        
        # pendulum
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([0.,0.,0.])
        
        [c1,s1] = self.trig( q )
        
        pts[1,0] =   s1 * self.l1
        pts[1,1] = - c1 * self.l1
        
        lines_pts.append( pts )
        lines_style.append( 'o-')
        lines_color.append( 'b' )
                
        return lines_pts , lines_style , lines_color
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """ 
        show torque as a rotating arrow around the joint
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        # Torque
        f         = u[0] # torque amplitude
        f_pos     = ( f > 0 )
        q         = x[0] - np.pi / 2  # rigid link angle
        max_angle = f * ( np.pi * 2 /3 / ( self.u_ub[0] ) )
        r         = self.l1 / 5.0  # radius of arc
        r1        = r/2            # length of arrows
        da        = 0.2            # angle discretization
        
        
        if abs(f) > (self.u_ub[0] * 0.05):
        
            if f_pos:
                angles = np.arange( 0, max_angle , da  ) + q 
            else:
                angles = np.arange( 0, max_angle * -1 , da  ) * -1 + q 
            n      = angles.size
            
            # Draw arc
            pts = np.zeros((n,3))
            for i , a in enumerate( angles ):
                pts[i,:] = [ r * np.cos(a) , r * np.sin(a) , 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            # Draw Arrow
            c = np.cos( max_angle + q )
            s = np.sin( max_angle + q )
            
            pts = np.zeros((3,3))
            pts[1,:] = [ r * c , r * s , 0 ]
            if f_pos:
                pts[0,:] = pts[1,:] + [ -r1/2*c+r1/2*s , -r1/2*s-r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c+r1/2*s , +r1/2*s-r1/2*c, 0 ]
            else:
                pts[0,:] = pts[1,:] + [ -r1/2*c-r1/2*s , -r1/2*s+r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c-r1/2*s , +r1/2*s+r1/2*c, 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        else:
            
            pts = np.zeros((3,3))
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
        
                
        return lines_pts , lines_style , lines_color
    
    
    
###############################################################################

class InvertedPendulum( SinglePendulum ):


    ############################
    def __init__(self):
        """ """
               
        # initialize standard params
        mechanical.MechanicalSystem.__init__(self, 1)
        
        # Name
        self.name = 'Inverted Pendulum'
        
        # params
        self.setparams()
        
        
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros( self.dof ) 
        
        [c1,s1] = self.trig( q )
        
        g[0] = -self.m1 * self.gravity * self.lc1 * s1

        return g
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        
        q2 = q + np.array([ np.pi ])
        
        return SinglePendulum.forward_kinematic_lines( self, q2)

    
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        
        x2 = x + np.array([  np.pi , 0  ])
        
        return SinglePendulum.forward_kinematic_lines_plus( self, x2 , u , t )
    
    
    
        
        
        
        
        
##############################################################################
        
class DoublePendulum( mechanical.MechanicalSystem ):
    """ 

    """
    
    ############################
    def __init__(self):
        """ """
               
        # initialize standard params
        mechanical.MechanicalSystem.__init__(self, 2 )
        
        # Name
        self.name = 'Double Pendulum'
        
        # params
        self.setparams()
        
        self.l_domain = 3
                
            
    #############################
    def setparams(self):
        """ Set model parameters here """
        
        self.l1  = 1 
        self.l2  = 1
        self.lc1 = 1
        self.lc2 = 1
        
        self.m1 = 1
        self.I1 = 0
        self.m2 = 1
        self.I2 = 0
        
        self.gravity = 9.81
        
        self.d1 = 0
        self.d2 = 0
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )
        c12 = np.cos( q[0] + q[1] )
        s12 = np.sin( q[0] + q[1] )
        
        return [c1,s1,c2,s2,c12,s12]
        
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        H = np.zeros((2,2))
        
        H[0,0] = self.m1 * self.lc1**2 + self.I1 + self.m2 * ( self.l1**2 + self.lc2**2 + 2 * self.l1 * self.lc2 * c2 ) + self.I2
        H[1,0] = self.m2 * self.lc2**2 + self.m2 * self.l1 * self.lc2 * c2 + self.I2
        H[0,1] = H[1,0]
        H[1,1] = self.m2 * self.lc2 ** 2 + self.I2
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        h = self.m2 * self.l1 * self.lc2 * s2
        
        C = np.zeros((2,2))
        
        C[0,0] = - h  * dq[1]
        C[1,0] =   h  * dq[0]
        C[0,1] = - h * ( dq[0] + dq[1] )
        C[1,1] = 0

        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.diag( np.ones( self.dof ) ) #  identity matrix
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        g1 = (self.m1 * self.lc1 + self.m2 * self.l1 ) * self.gravity
        g2 = self.m2 * self.lc2 * self.gravity
        
        G = np.zeros(2)
        
        G[0] = - g1 * s1 - g2 * s12
        G[1] = - g2 * s12

        return G
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        D = np.zeros((2,2))
        
        D[0,0] = self.d1
        D[1,0] = 0
        D[0,1] = 0
        D[1,1] = self.d2
        
        d = np.dot( D , dq )
        
        return d
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = self.l_domain
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###############################
        # ground line
        ###############################
        
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([-10,0,0])
        pts[1,:] = np.array([+10,0,0])
        
        lines_pts.append( pts )
        lines_style.append( '--' )
        lines_color.append( 'k' )
        
        ###########################
        # pendulum kinematic
        ###########################
        
        pts      = np.zeros(( 3 , 3 ))
        pts[0,:] = np.array([0,0,0])
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        pts[1,0] = self.l1 * s1
        pts[1,1] = self.l1 * c1
        
        pts[2,0] = self.l1 * s1 + self.l2 * s12
        pts[2,1] = self.l1 * c1 + self.l2 * c12
        
        lines_pts.append( pts )
        lines_style.append( 'o-' )
        lines_color.append( 'b' )
                
        return lines_pts , lines_style , lines_color
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """ 
        show torques as a rotating arrow around the joint
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        r          = self.l1 / 5.0  # radius of arc
        r1         = r/2            # length of arrows
        da         = 0.2            # angle discretization
        
        # Torque 1
        x1         = 0
        y1         = 0
        f1         = -u[0] # torque amplitude
        f1_pos     = ( f1 > 0 )
        q1         = -x[0] +  np.pi /2     # rigid link angle
        max_angle1 = f1 * ( np.pi * 2 /3 / ( self.u_ub[0] ) )
        
        if abs(f1) > (self.u_ub[0] * 0.05):
        
            if f1_pos:
                angles = np.arange( 0, max_angle1 , da  ) + q1 
            else:
                angles = np.arange( 0, max_angle1 * -1 , da  ) * -1 + q1 
            n      = angles.size
            
            # Draw arc
            pts = np.zeros((n,3))
            for i , a in enumerate( angles ):
                pts[i,:] = [ r * np.cos(a) + x1 , r * np.sin(a) + y1 , 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-' )
            lines_color.append( 'r' )
            
            # Draw Arrow
            a = max_angle1 + q1 
            c = np.cos( a )
            s = np.sin( a )
            
            pts = np.zeros((3,3))
    
            pts[1,:] =  [ r * c + x1 , r * s + y1 , 0 ]
            if f1_pos:
                pts[0,:] = pts[1,:] + [ -r1/2*c+r1/2*s , -r1/2*s-r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c+r1/2*s , +r1/2*s-r1/2*c, 0 ]
            else:
                pts[0,:] = pts[1,:] + [ -r1/2*c-r1/2*s , -r1/2*s+r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c-r1/2*s , +r1/2*s+r1/2*c, 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        else:
            
            pts = np.zeros((3,3))
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        
        # Torque 2
        x2         = self.l1 * np.sin( -q1 + np.pi /2 )
        y2         = self.l1 * np.cos( -q1 + np.pi /2)
        f2         = -u[1] # torque 2 amplitude
        f2_pos     = ( f2 > 0 )
        q12        = - x[0] - x[1] +  np.pi /2  # rigid link angle
        max_angle2 = f2 * ( np.pi * 2 /3 / ( self.u_ub[1] ) )
        
        
        if abs(f2) > (self.u_ub[0] * 0.05):
        
            if f2_pos:
                angles = np.arange( 0, max_angle2 , da  ) + q12 
            else:
                angles = np.arange( 0, max_angle2 * -1 , da  ) * -1 + q12 
            n      = angles.size
            
            # Draw arc
            pts = np.zeros((n,3))
            for i , a in enumerate( angles ):
                pts[i,:] = [ r * np.cos(a) + x2 , r * np.sin(a) + y2 , 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-' )
            lines_color.append( 'r' )
            
            # Draw Arrow
            a = max_angle2 + q12 
            c = np.cos( a )
            s = np.sin( a )
            
            pts = np.zeros((3,3))
    
            pts[1,:] =  [ r * c + x2 , r * s + y2 , 0 ]
            if f2_pos:
                pts[0,:] = pts[1,:] + [ -r1/2*c+r1/2*s , -r1/2*s-r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c+r1/2*s , +r1/2*s-r1/2*c, 0 ]
            else:
                pts[0,:] = pts[1,:] + [ -r1/2*c-r1/2*s , -r1/2*s+r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c-r1/2*s , +r1/2*s+r1/2*c, 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        else:
            
            pts = np.zeros((3,3))
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        
                
        return lines_pts , lines_style , lines_color




##############################################################################
        
class Acrobot( DoublePendulum ):
    """ 
    Double pendulum with a single motor at the elbow

    """
    
    ############################
    def __init__(self):
        """ """
               
        # initialize standard params
        mechanical.MechanicalSystem.__init__(self, dof=2, actuators=1)
        
        self.name = 'Acrobot'
        
        self.input_label[0] = 'tau'
        self.input_units[0] = '[Nm]'
        
        self.u_lb[0] = -10
        self.u_ub[0] = +10
        
        # params
        
        self.l1  = 1 
        self.l2  = 1
        self.lc1 = 1
        self.lc2 = 1
        
        self.m1 = 1
        self.I1 = 0
        self.m2 = 1
        self.I2 = 0
        
        self.gravity = 9.81
        
        self.d1 = 0
        self.d2 = 0
        
        self.l_domain = 3
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.array([[0],[1]])
        
        return B
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """ 
        show torques as a rotating arrow around the joint
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        r          = self.l1 / 5.0  # radius of arc
        r1         = r/2            # length of arrows
        da         = 0.2            # angle discretization
        
        q1         = -x[0] +  np.pi /2     # rigid link angle
            
        
        # Torque 2
        x2         = self.l1 * np.sin( -q1 + np.pi /2 )
        y2         = self.l1 * np.cos( -q1 + np.pi /2)
        f2         = -u[0] # torque 2 amplitude
        f2_pos     = ( f2 > 0 )
        q12        = - x[0] - x[1] +  np.pi /2  # rigid link angle
        max_angle2 = f2 * ( np.pi * 2 /3 / ( self.u_ub[0] ) )
        
        
        if abs(f2) > (self.u_ub[0] * 0.05):
        
            if f2_pos:
                angles = np.arange( 0, max_angle2 , da  ) + q12 
            else:
                angles = np.arange( 0, max_angle2 * -1 , da  ) * -1 + q12 
            n      = angles.size
            
            # Draw arc
            pts = np.zeros((n,3))
            for i , a in enumerate( angles ):
                pts[i,:] = [ r * np.cos(a) + x2 , r * np.sin(a) + y2 , 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-' )
            lines_color.append( 'r' )
            
            # Draw Arrow
            a = max_angle2 + q12 
            c = np.cos( a )
            s = np.sin( a )
            
            pts = np.zeros((3,3))
    
            pts[1,:] =  [ r * c + x2 , r * s + y2 , 0 ]
            if f2_pos:
                pts[0,:] = pts[1,:] + [ -r1/2*c+r1/2*s , -r1/2*s-r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c+r1/2*s , +r1/2*s-r1/2*c, 0 ]
            else:
                pts[0,:] = pts[1,:] + [ -r1/2*c-r1/2*s , -r1/2*s+r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c-r1/2*s , +r1/2*s+r1/2*c, 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        else:
            
            pts = np.zeros((3,3))
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        
                
        return lines_pts , lines_style , lines_color
    

    


###############################################################################
# Two independent Pendulum
###############################################################################

class TwoIndependentSinglePendulum( mechanical.MechanicalSystem ):
    """ Two Pendulum with a point mass and inertial rod.

    """

    ############################
    def __init__(self):
        """ """
               
        # initialize standard params
        mechanical.MechanicalSystem.__init__(self, 2)
        
        # Name
        self.name = 'Two Independent Single Pendulum'
        
        # params
        
        # kinematic
        self.l1  = 2 
        self.lc1 = 1
        
        # dynamic
        self.m1       = 1
        self.I1       = 1
        self.gravity  = 9.81
        self.d1       = 0
        
        
    ##############################
    def trig(self, q ):
        """ Compute cos and sin """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )

        return [c1,s1,c2,s2]
    
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        H = np.zeros((self.dof,self.dof))
        
        H[0,0] = self.m1 * self.lc1**2 + self.I1
        H[1,1] = self.m1 * self.lc1**2 + self.I1
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        C = np.zeros((self.dof,self.dof))

        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.diag( np.ones( self.dof ) ) #  identity matrix
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros( self.dof ) 
        
        [c1,s1,c2,s2] = self.trig( q )
        
        g[0] = self.m1 * self.gravity * self.lc1 * s1
        g[1] = self.m1 * self.gravity * self.lc1 * s2

        return g
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d    = np.zeros( self.dof ) 
        
        d[0] = self.d1 * dq[0]
        d[1] = self.d1 * dq[1]
        
        return d
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = 5
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        
        offset = 2
        
        # ground line
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([-10,0,0])
        pts[1,:] = np.array([+10,0,0])
        
        lines_pts.append( pts )
        
        # pendulum no 1
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([- offset,0,0])
        
        [c1,s1,c2,s2] = self.trig( q )
        
        pts[1,:] = np.array([ s1 * self.l1 - offset , - c1 * self.l1 ,0])
        
        lines_pts.append( pts )
        
        # pendulum no 1
        
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([offset,0,0])
        
        [c1,s1,c2,s2] = self.trig( q )
        
        pts[1,:] = np.array([ s2 * self.l1 + offset , - c2 * self.l1 ,0])
        
        lines_pts.append( pts )
                
        return lines_pts
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """ 
        show torque as a rotating arrow around the joint
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        # FIrst pendulum
        
        # Torque
        f         = u[0] # torque amplitude
        f_pos     = ( f > 0 )
        q         = x[0] - np.pi / 2  # rigid link angle
        max_angle = f * ( np.pi * 2 /3 / ( self.u_ub[0] ) )
        r         = self.l1 / 5.0  # radius of arc
        r1        = r/2            # length of arrows
        da        = 0.2            # angle discretization
        
        
        if abs(f) > (self.u_ub[0] * 0.05):
        
            if f_pos:
                angles = np.arange( 0, max_angle , da  ) + q 
            else:
                angles = np.arange( 0, max_angle * -1 , da  ) * -1 + q 
            n      = angles.size
            
            # Draw arc
            pts = np.zeros((n,3))
            for i , a in enumerate( angles ):
                pts[i,:] = [ r * np.cos(a) - 2.0, r * np.sin(a) , 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            # Draw Arrow
            c = np.cos( max_angle + q )
            s = np.sin( max_angle + q )
            
            pts = np.zeros((3,3))
            pts[1,:] = [ r * c - 2.0 , r * s , 0 ]
            if f_pos:
                pts[0,:] = pts[1,:] + [ -r1/2*c+r1/2*s , -r1/2*s-r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c+r1/2*s , +r1/2*s-r1/2*c, 0 ]
            else:
                pts[0,:] = pts[1,:] + [ -r1/2*c-r1/2*s , -r1/2*s+r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c-r1/2*s , +r1/2*s+r1/2*c, 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        else:
            
            pts = np.zeros((3,3))
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        # Second pendulum
        
        # Torque
        f         = u[1] # torque amplitude
        f_pos     = ( f > 0 )
        q         = x[1] - np.pi / 2  # rigid link angle
        max_angle = f * ( np.pi * 2 /3 / ( self.u_ub[1] ) )
        r         = self.l1 / 5.0  # radius of arc
        r1        = r/2            # length of arrows
        da        = 0.2            # angle discretization
        
        
        if abs(f) > (self.u_ub[1] * 0.05):
        
            if f_pos:
                angles = np.arange( 0, max_angle , da  ) + q 
            else:
                angles = np.arange( 0, max_angle * -1 , da  ) * -1 + q 
            n      = angles.size
            
            # Draw arc
            pts = np.zeros((n,3))
            for i , a in enumerate( angles ):
                pts[i,:] = [ r * np.cos(a) + 2.0 , r * np.sin(a) , 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            # Draw Arrow
            c = np.cos( max_angle + q )
            s = np.sin( max_angle + q )
            
            pts = np.zeros((3,3))
            pts[1,:] = [ r * c + 2.0 , r * s , 0 ]
            if f_pos:
                pts[0,:] = pts[1,:] + [ -r1/2*c+r1/2*s , -r1/2*s-r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c+r1/2*s , +r1/2*s-r1/2*c, 0 ]
            else:
                pts[0,:] = pts[1,:] + [ -r1/2*c-r1/2*s , -r1/2*s+r1/2*c, 0 ]
                pts[2,:] = pts[1,:] + [ +r1/2*c-r1/2*s , +r1/2*s+r1/2*c, 0 ]
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
        else:
            
            pts = np.zeros((3,3))
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
            
            lines_pts.append( pts )
            lines_style.append( '-')
            lines_color.append( 'r' )
        
                
        return lines_pts , lines_style , lines_color
        
        
        
        
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    
    if False:
    
        sys = TwoIndependentSinglePendulum()
        
        def t2u(t):
            return np.array([ t + sys.u_lb[0] , t + sys.u_lb[1]])
        
        sys.t2u   = t2u
        sys.ubar[0] = 2
        sys.ubar[1] = 1
        sys.x0[0] = 3.14
        sys.compute_trajectory( 10 )
        sys.plot_trajectory('xu')
        sys.animate_simulation()
        
    if False:
        
        sys = Acrobot()
        
        def t2u(t):
            return np.array([ t -2 ])
        
        sys.t2u   = t2u
        sys.x0[0] = 3.14
        sys.compute_trajectory( 10 )
        sys.plot_trajectory('xu')
        sys.animate_simulation()
        
        
    if True:
        
        #sys = SinglePendulum()
        
        sys = InvertedPendulum()
        
        def t2u(t):
            return np.array([ 5 * np.sin(3*t) ])
        
        sys.t2u   = t2u
        sys.x0[0] = 3.14
        sys.compute_trajectory( 10 )
        sys.plot_trajectory('xu')
        sys.animate_simulation()
        
        
    
        


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:13:40 2018

@author: Alexandre
"""

from pyro.dynamic  import equation

sys      = equation.VanderPol()

sys.x0[1] = 4.0
sys.compute_trajectory( tf = 20 )
sys.plot_phase_plane_trajectory()
#sys.plot_trajectory()

sys.x0[1] = 0.1
sys.compute_trajectory( tf = 20 )
sys.plot_phase_plane_trajectory()


# -*- coding: utf-8 -*-

###############################################################################
import numpy as np
import matplotlib.pyplot as plt


from pyro.analysis import graphical
from pyro.planning import trajectorygeneration
from pyro.dynamic import rigidbody
from pyro.planning import plan

###############################################################################

# fixed initial position for now
# initial angular velocity is related to jerk of trajectory
x0 = 0.0
y0 = 0.0
z0 = 0.0

# fixed final position for now
xf = 10.0
yf = 10.0
zf = np.pi 
tf = 5

ddx0 = xf * 0.1
ddy0 = 0.0# ddx0 * np.tan(z0)

ddxf = 0.0
ddyf = yf * -0.1#-ddxf * np.tan(zf)


gex = trajectorygeneration.SingleAxisPolynomialTrajectoryGenerator(poly_N=9)
gex.x0_N = 3
gex.xf_N = 3
gex.x0 = np.array([x0, 0, ddx0, 0, 0, 0])
gex.xf = np.array([xf, 0, ddxf, 0, 0, 0])
gex.poly_N = 7
gex.Ws = np.array([0.01, 1.0, 1.0, 1.0, 1.0, 1.0, .0])
px, X, t = gex.solve()
x = X[0, :]
dx = X[1, :]
ax = X[2, :]
dax = X[3, :]
ddax = X[4, :]

gey = trajectorygeneration.SingleAxisPolynomialTrajectoryGenerator(poly_N=9)
gey.x0_N = 3
gey.xf_N = 3
gey.x0 = np.array([y0, 0, ddy0, 0, 0, 0])
gey.xf = np.array([yf, 0, ddyf, 0, 0, 0])
gey.poly_N = 7
gey.Ws = np.array([0.01, 1.0, 1.0, 1.0, 1.0, 1.0, .0])
py, Y, t = gey.solve()
y = Y[0, :]
dy = Y[1, :]
ay = Y[2, :]
day = Y[3, :]
dday = Y[4, :]

# Position theta
theta = np.arctan2(ay, ax)
# theta = np.arctan( (ay/ax))
s = np.sin(theta)
c = np.cos(theta)
dtheta = (day * c - dax * s) / (ay * s + ax * c)  # TODO check analytical equation, seems wrong
ddtheta = (
    s * (-ddax + ax * dtheta**2 - 2 * day * dtheta)
    + c * (dday - ay * dtheta**2 - 2 * dax * dtheta)
) / (
    ax * c + ay * s
)  # TODO check analytical equation, seems wrong

dtheta_num = np.diff(theta, n=1, prepend=0.0)
ddtheta_num = np.diff(dtheta, n=1, prepend=0.0)

# dtheta = dtheta_num
# ddtheta = ddtheta_num

# Create traj
steps = len(t)
xs = np.zeros((steps, 6))
ys = np.zeros((steps, 6))
us = np.zeros((steps, 2))
dxs = np.zeros((steps, 6))

sys = rigidbody.RigidBody2D()

sys.mass = 0.8
sys.inertia = 1.0
sys.l_t = 1.0

m = sys.mass
J = sys.inertia
r = sys.l_t

x_cg = x - J / (m * r) * np.cos(theta)
y_cg = y - J / (m * r) * np.sin(theta)

xs[:, 0] = x_cg
xs[:, 1] = y_cg
xs[:, 2] = theta

M = np.array([[m, 0], [0, m]])

ax_cg = (
    ax + J / (m * r) * np.sin(theta) * ddtheta + J / (m * r) * np.cos(theta) * dtheta**2
)
ay_cg = (
    ay - J / (m * r) * np.cos(theta) * ddtheta + J / (m * r) * np.sin(theta) * dtheta**2
)

# COmpute forces
for i in range(steps):
    R = np.array(
        [[np.cos(theta[i]), -np.sin(theta[i])], [np.sin(theta[i]), np.cos(theta[i])]]
    )
    a_cg = np.array([ax_cg[i], ay_cg[i]])
    us[i, :] = np.linalg.inv(R) @ M @ a_cg


traj = plan.Trajectory(xs, us, t, dxs, ys)

fig, axes = plt.subplots(
    2, figsize=graphical.default_figsize, dpi=graphical.default_dpi, frameon=True
)

axes[0].plot(t, ax_cg, "b")
axes[0].set_ylabel("ax_cg", fontsize=graphical.default_fontsize)
axes[0].set_xlabel("t", fontsize=graphical.default_fontsize)
axes[0].tick_params(labelsize=graphical.default_fontsize)
axes[0].grid(True)

axes[1].plot(t, ay_cg, "b")
axes[1].set_ylabel("ay_cg", fontsize=graphical.default_fontsize)
axes[1].set_xlabel("t", fontsize=graphical.default_fontsize)
axes[1].tick_params(labelsize=graphical.default_fontsize)
axes[1].grid(True)


fig, axes = plt.subplots(
    3, figsize=graphical.default_figsize, dpi=graphical.default_dpi, frameon=True
)

axes[0].plot(t, theta, "b")
axes[0].set_ylabel("Theta", fontsize=graphical.default_fontsize)
axes[0].set_xlabel("v", fontsize=graphical.default_fontsize)
axes[0].tick_params(labelsize=graphical.default_fontsize)
axes[0].grid(True)

axes[1].plot(t, dtheta, "b")
# axes[1].plot(t, dtheta_num, "r")
axes[1].set_ylabel("w", fontsize=graphical.default_fontsize)
axes[1].set_xlabel("t", fontsize=graphical.default_fontsize)
axes[1].tick_params(labelsize=graphical.default_fontsize)
axes[1].grid(True)

axes[2].plot(t, ddtheta, "b")
# axes[2].plot(t, ddtheta_num, "r")
axes[2].set_ylabel("dw", fontsize=graphical.default_fontsize)
axes[2].set_xlabel("t", fontsize=graphical.default_fontsize)
axes[2].tick_params(labelsize=graphical.default_fontsize)
axes[2].grid(True)

fig.tight_layout()
fig.canvas.draw()

plt.show()


fig, axes = plt.subplots(
    1, figsize=graphical.default_figsize, dpi=graphical.default_dpi, frameon=True
)

axes.plot(x, y, "r")
axes.plot(x_cg, y_cg, "b")
axes.set_ylabel("y", fontsize=graphical.default_fontsize)
axes.set_xlabel("x", fontsize=graphical.default_fontsize)
axes.axis("equal")
axes.set(xlim=(-5, 25), ylim=(-5, 25))
axes.tick_params(labelsize=graphical.default_fontsize)
axes.grid(True)

fig.tight_layout()
fig.canvas.draw()

plt.show()


sys.traj = traj

ani = sys.animate_simulation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 14 23:19:16 2020
@author: alex
------------------------------------
Fichier d'amorce pour le controleur de drillage
"""

import numpy as np

from pyro.control.robotcontrollers import EndEffectorPD


class CustomDrillingController( EndEffectorPD ) :
    """ 

    """
    
    ############################
    def __init__(self, robot_model ):
        """ """
        
        super().__init__(robot_model)
        
        # Label
        self.name = 'Custom Drilling Controller'
        
        """
        ###################################################
        # Vos paramètres de loi de commande ci-dessous !!
        ###################################################
        """
        
        # Target effector force
        self.rbar = np.array([0,0,0]) 
        
    
    #############################
    def c( self , y , r , t = 0 ):

        f_e = r  # Reference de force cartésienne cible
        x   = y  # Feedback from sensors (full state feedback)
        
        [ q , dq ] = self.x2q( x ) # extraction of joint position and speed
        
        
        """
        ##################################
        
        
        INPUTS
        f_e  : target force vector     3 x 1
        q    : joint positions vector  3 x 1
        dq   : joint velocity vector   3 x 1
        
        PRECOMPUTED FOR YOU
        J    : Jacobian matrix         3 x 3
        r    : effector position       3 x 1
        dr   : effector velocity       3 x 1
        
        OUPUTS
        tau  : joint torque command    3 x 1
        
        
        # Votre loi de commande ci-dessous !!!
        
        ##################################
        """
        
        J  = self.J( q ) # Jacobian computation
        r  = self.fwd_kin(q)  # cinématique directe
        dr = np.dot( J , dq ) # cinématique différentielle
        
        #TODO
                
        tau = np.zeros(self.m)  # place-holder de bonne dimension

        
        return tau
        


# -*- coding: utf-8 -*-
"""
Created on Sun Mar  6 15:09:12 2016

@author: alex
"""
###############################################################################
import numpy as np
import matplotlib.pyplot as plt

###############################################################################
from pyro.planning import plan
from pyro.analysis import simulation
from pyro.planning import filters


###############################################################################
class Node:
    """ node of the random tree """
    
    ############################
    def __init__(self, x , u , t , parent ):
        
        self.x = x  # Node coordinates in the state space
        self.u = u  # Control inputs used to get there
        self.t = t  # Time when arriving at x
        self.parent = parent # Previous node
        
    
    ############################
    def distanceTo(self, x_other ):
        """ Compute distance to otherNode """
        
        return np.linalg.norm( self.x - x_other )
        
        
###############################################################################
class RRT( plan.Planner ):
    """ Rapid Random Trees search algorithm """
    
    ############################
    def __init__(self, sys , x_start = None , x_goal = None ):
        
        # Set sys, default cost function x_start and x_goal
        plan.Planner.__init__(self, sys)
        
        # Default x_start is sys.x0
        if x_start is not None:
            self.x_start = x_start
          
        # Default x_goal is sys.xbar
        if x_goal is not None:
            self.x_goal = x_goal
        
        
        # Params
        self.dt                   = 0.1
        self.INF                  = 10000
        self.eps                  = 0.001
        self.steps                = 1
        self.goal_radius          = 0.5        
        self.alpha                = 0.9    # prob of random exploration
        self.beta                 = 0.0    # prob of random u
        self.max_nodes            = 2000   # maximum number of nodes
        self.max_distance_compute = 2000   # max  nodes to check distance
        self.max_solution_time    = 100    # won"t look for longuer solution 
        
        self.test_u_domain        = False  # run a check on u input 
                
        # Ploting
        self.dyna_plot            = True
        self.dyna_node_no_update  = 100
        self.fontsize             = 5
        self.figsize              = (3, 2)
        self.dpi                  = 300
        self.x_axis               = 0  # State to plot on x axis
        self.y_axis               = 1  # State to plot on y axis
        self.z_axis               = 2  # State to plot on z axis
        
        # Debuging mode
        self.debug = False
        
        # Init
        self.discretizeactions()
        self.solution_is_found     = False
        self.randomized_input      = False
        
        
    #############################
    def discretizeactions(self, n = 3 ):
        """ generate the list of possible control inputs """
        
        self.u_options = [ self.sys.u_lb ,  self.sys.ubar , self.sys.u_ub ]
        
        
    #############################
    def compute_solution(self, x_goal = None ):
    
        self.find_path_to_goal( x_goal )
        
        
    ############################
    def rand_state(self):    
        """ Sample a random state """
        
        ranges = self.sys.x_ub - self.sys.x_lb
        
        x_random = np.random.rand( self.sys.n ) * ranges + self.sys.x_lb
        
        #if not( self.sys.isavalidstate( x_random ) ):
                # Sample again (recursivity)
                #x_random = self.rand_state()
        
        return x_random
        
    
    ############################
    def rand_input(self, x = 0 ):    
        """ Sample a random input """
        
        # random selection
        n_options = len( self.u_options )
        j         = np.random.randint( 0 , n_options )
        u         = self.u_options[j]
        
        # if u domain check is active
        if self.test_u_domain :
            
            # If new sample is not a valid option
            if not( self.sys.isavalidinput( x , u ) ):
                
                # Sample again (recursivity)
                u = self.rand_input( x )
        
        return u
        
        
    ############################
    def nearest_neighbor(self, x_target ):    
        """ Get the nearest node to a given state x """
        
        closest_node = None
        min_distance = self.INF
        
        # Number of nodes is small
        if len(self.nodes) < self.max_distance_compute + 1 :
            # Brute force        
            for node in self.nodes:
                d = node.distanceTo( x_target )
                if d < min_distance:
                    if node.t < self.max_solution_time:
                        min_distance = d
                        closest_node = node
        
        # Else if there is too many nodes to check
        # TODO: fix this, Implement efficient nearest search?
        else:
            # Check only last X nodes
            for i in range(self.max_distance_compute):
                node = self.nodes[ len(self.nodes) - i - 1 ]
                d = node.distanceTo( x_target )
                if d < min_distance:
                    if node.t < self.max_solution_time:
                        min_distance = d
                        closest_node = node
        
        return closest_node
        
        
    ############################
    def select_control_input(self, x_target , closest_node ):    
        """ pick control input """
        
        # Select a random control input
        if self.randomized_input :
            
            u          = self.rand_input( closest_node.x )
            x_next     = self.sys.x_next( closest_node.x , 
                                          u , 
                                          closest_node.t , 
                                          self.dt ,
                                          self.steps 
                                          )
            
            t_next     = closest_node.t + self.dt * self.steps
            new_node   = Node( x_next , u , t_next  , closest_node )
            
            if not( self.sys.isavalidstate( x_next ) ):
                new_node = None
        
        # Pick control input that bring the sys close to random point
        else:
            
            new_node     = None
            min_distance = self.INF
            
            for u in self.u_options:
                
                # if u domain check is active
                if self.test_u_domain:
                    # if input is not valid
                    if not( self.sys.isavalidinput( closest_node.x , u ) ):
                        # Skip this u
                        continue
                
                x_next     = self.sys.x_next( closest_node.x , 
                                              u , 
                                              closest_node.t , 
                                              self.dt ,
                                              self.steps 
                                              )
                
                t_next     = closest_node.t + self.dt * self.steps
                node       = Node( x_next , u , t_next  , closest_node )
                
                d = node.distanceTo( x_target )
                
                if ( d < min_distance ) and self.sys.isavalidstate( x_next ) :
                    min_distance = d
                    new_node     = node
                    
                
        return new_node
    
    
    ############################
    def init_tree(self):    
        """ """
        # Init Tree
        self.start_node = Node( self.x_start , None , 0  , None )
        self.nodes = []
        self.nodes.append( self.start_node )
        
    
    ############################
    def one_step(self):    
        """ """
        x_random  = self.rand_state()
        
        node_near = self.nearest_neighbor( x_random )
        
        # if a valid neighbor was found
        if not node_near == None:
            new_node  = self.select_control_input( x_random , node_near )
            
            # if there is a valid control input
            if not new_node == None:
                self.nodes.append( new_node )
        
        
    ############################
    def compute_steps(self , n , plot = False ):    
        """ """
        for i in range( n ):
            self.one_step()
            
        if plot:
            self.plot_tree()
           
        
    ############################
    def find_path_to_goal(self, x_goal = None ):
        """ """
        
        # Default x_goal is sys.xbar
        if x_goal is not None:
            self.x_goal = x_goal
            
        self.init_tree()
        
        # Variables
        succes   = False
        no_nodes = 0
        
        # Plot
        if self.dyna_plot:
            self.dyna_plot_init()
        
        # Explore the tree
        while not succes:
            
            # Exploration:
            if np.random.rand() > self.alpha :
                # Try to converge to goal
                x_random = self.x_goal
                self.randomized_input = False
                
            else:
                # Random exploration
                x_random  = self.rand_state()
                
                # self.beta = probability of random exploration
                self.randomized_input = ( np.random.rand() < self.beta )

            node_near = self.nearest_neighbor( x_random )
            
            # if a valid neighbor was found
            if not node_near == None:
                new_node  = self.select_control_input( x_random , node_near )
                
                # if there is a valid control input
                if not new_node == None:
                    self.nodes.append( new_node )
            
                    # Distance to goal
                    d = new_node.distanceTo( self.x_goal )
                    
                    no_nodes = no_nodes + 1
                    
                    ##################################################
                    # Debug
                    if self.debug:
                        print(x_random, node_near.x , new_node.x )
                        wait = input("PRESS ENTER TO CONTINUE.")
                    ###################################################
                    
                    # Plot
                    if self.dyna_plot:
                        self.dyna_plot_add_node( new_node , no_nodes )
                    
                    # Succes?
                    if d < self.goal_radius:
                        succes = True
                        self.goal_node = new_node
                else:
                    pass
                    #print('on obstacle')
                    
                
            # Tree reset
            if no_nodes == self.max_nodes:
                
                print('\n-----------------------------------------------',
                      '\nRRT reseting tree',
                      '\n-----------------------------------------------')
                no_nodes = 0
                self.nodes = []
                self.nodes.append( self.start_node )
                
                if self.dyna_plot :
                    self.dyna_plot_clear()
        
        print('\n-----------------------------------------------',
              '\nRRT found a path to the goal',
              '\n-----------------------------------------------')
        
        # Compute Path
        self.path_to_traj()
        
        # Plot
        if self.dyna_plot:
            self.dyna_plot_solution()
        
                
    ############################
    def path_to_traj(self):
        """ 
        Extract the solution in the form of a trajectory
        
        """
        
        node = self.goal_node
        
        t      = 0
        
        x_list  = []
        u_list  = []
        t_list  = []
        dx_list = []
        
        self.path_node_list = []
        
        # Until node = start_node
        while node.distanceTo( self.x_start ) > self.eps:
            
            self.path_node_list.append( node )
            
            x_list.append( node.parent.x   )
            u_list.append( node.u     )
            t_list.append( node.parent.t   )
            
            dx_list.append( self.sys.f( node.parent.x , 
                                        node.u , 
                                        node.parent.t )  ) # state derivative
            
            # Previous Node
            node  = node.parent 

        # Arrange Time array
        t = np.array( t_list )
        t = np.flipud( t )
        
        # Arrange Input array
        u = np.array( u_list ).T
        u = np.fliplr( u )
        
        # Arrange State array
        x = np.array( x_list ).T
        x = np.fliplr( x )
        
        # Arrange State Derivative array
        dx = np.array( dx_list ).T
        dx = np.fliplr( dx )
            
        # Save solution
        self.traj = simulation.Trajectory(x.T, u.T, t.T, dx.T, x.T)
        
        # Create open-loop controller
        self.open_loop_controller = plan.OpenLoopController( self.traj )
        
        #
        self.solution_is_found = True
        
    
    ############################
    def low_pass_solution(self, fc = 1 ):
        """ Fc = cutoff freq in Hz """
        
        #Memorize original raw trajectory plan
        self.raw_trajectory_solution = self.traj
        
        tf = filters.TrajectoryFilter( fc , self.dt )
        
        # Low pass filter
        filtered_traj = tf.low_pass_filter_traj( self.traj)
        
        self.traj = filtered_traj


    ##################################################################
    ### Ploting functions
    ##################################################################            
                
    ############################
    def plot_tree(self):
        """ """
        
        # Create figure
        self.fig_tree = plt.figure(figsize=(3, 2), dpi=300, frameon=True)
        
        # Set window title
        self.fig_tree.canvas.manager.set_window_title('RRT tree search for ' + 
                                            self.sys.name )
        
        # Create axe
        ax       = self.fig_tree.add_subplot(111)
        
        # Plot Tree
        for node in self.nodes:
            if not(node.parent==None):
                ax.plot( 
                [node.x[ self.x_axis ],node.parent.x[ self.x_axis ]] , 
                [node.x[ self.y_axis ],node.parent.x[ self.y_axis ]] , 'o-')
        
        # Plot Solution Path
        if self.solution_is_found:
            for node in self.path_node_list:
                if not( node.parent == None ):
                    ax.plot( 
                    [node.x[ self.x_axis ],node.parent.x[ self.x_axis ]] , 
                    [node.x[ self.y_axis ],node.parent.x[ self.y_axis ]] , 'r')
        
        # Set axis labels
        ax.set_xlabel(
                self.sys.state_label[ self.x_axis ] + ' ' +
                self.sys.state_units[ self.x_axis ] , fontsize=self.fontsize)
        ax.set_ylabel(
                self.sys.state_label[ self.y_axis ] + ' ' + 
                self.sys.state_units[ self.y_axis ] , fontsize=self.fontsize)
        
        # Set domain
        ax.set_xlim(
                [ self.sys.x_lb[ self.x_axis ] , self.sys.x_ub[ self.x_axis ] ]
                )
        ax.set_ylim(
                [ self.sys.x_lb[ self.y_axis ] , self.sys.x_ub[ self.y_axis ] ]
                )
        
        #Set grid
        ax.grid(True)
        ax.tick_params( labelsize = self.fontsize )
        
        self.fig_tree.tight_layout()
        self.fig_tree.show()
        
        return ax
        
        
    ############################
    def plot_tree_3d(self):
        """ """
        
        # Create figure
        self.fig_tree_3d = plt.figure( figsize = self.figsize, dpi = self.dpi )
        
        # Set window title
        self.fig_tree_3d.canvas.manager.set_window_title('RRT tree search for ' + 
                                            self.sys.name )
        
        # Create Axe
        ax = self.fig_tree_3d.add_subplot(projection='3d')
        
        # Plot Tree
        for node in self.nodes:
            if not( node.parent == None ):
                ax.plot( 
                [ node.x[ self.x_axis ] , node.parent.x[ self.x_axis ]] ,
                [ node.x[ self.y_axis ] , node.parent.x[ self.y_axis ]] ,
                [ node.x[ self.z_axis ] , node.parent.x[ self.z_axis ]] , 'o-')
        
        # Plot Solution Path
        if not self.solution_is_found == None:
            for node in self.path_node_list:
                if not( node.parent == None ):
                    ax.plot( 
                    [ node.x[ self.x_axis ] , node.parent.x[ self.x_axis ]] ,
                    [ node.x[ self.y_axis ] , node.parent.x[ self.y_axis ]] ,
                    [ node.x[ self.z_axis ] , node.parent.x[ self.z_axis ]] , 
                    'r')
        
        # Set domain
        ax.set_xlim3d( [ self.sys.x_lb[ self.x_axis ] ,
                         self.sys.x_ub[ self.x_axis ] ] )
        ax.set_ylim3d( [ self.sys.x_lb[ self.y_axis ] , 
                         self.sys.x_ub[ self.y_axis ] ]  )
        ax.set_zlim3d( [ self.sys.x_lb[ self.z_axis ] , 
                         self.sys.x_ub[ self.z_axis ] ]  )
    
        # Set labels
        ax.set_xlabel(
        self.sys.state_label[ self.x_axis ] + ' ' + 
        self.sys.state_units[ self.x_axis ] ,  fontsize=self.fontsize )
        
        ax.set_ylabel(
        self.sys.state_label[ self.y_axis ] + ' ' +
        self.sys.state_units[ self.y_axis ] , fontsize=self.fontsize)
        
        ax.set_zlabel(
        self.sys.state_label[ self.z_axis ] + ' ' + 
        self.sys.state_units[ self.z_axis ] , fontsize=self.fontsize)
        
        # Grid
        ax.grid(True)
        ax.tick_params( labelsize = self.fontsize )
        
        self.fig_tree_3d.tight_layout()
        self.fig_tree_3d.show()
        
        return ax
        
    
    ############################
    def dyna_plot_init(self):
        
        # Create figure
        self.fig_tree_dyna = plt.figure(figsize=(3, 2),dpi=300, frameon=True)
        
        # Set window title
        self.fig_tree_dyna.canvas.manager.set_window_title('RRT tree search for ' + 
                                            self.sys.name )
        
        # Create axe
        ax  = self.fig_tree_dyna.add_subplot(111)
        
        self.time_template = 'Number of nodes = %i'
        self.time_text = ax.text(
                0.05, 0.05, '', transform=ax.transAxes, 
                fontsize=self.fontsize )
        
        # Set domain
        ax.set_xlim( [ self.sys.x_lb[ self.x_axis ] ,
                       self.sys.x_ub[ self.x_axis ] ] )
        ax.set_ylim( [ self.sys.x_lb[ self.y_axis ] , 
                       self.sys.x_ub[ self.y_axis ] ]  )
    
        # Set labels
        ax.set_xlabel(
        self.sys.state_label[ self.x_axis ] + ' ' + 
        self.sys.state_units[ self.x_axis ] ,  fontsize=self.fontsize )
        
        ax.set_ylabel(
        self.sys.state_label[ self.y_axis ] + ' ' +
        self.sys.state_units[ self.y_axis ] , fontsize=self.fontsize)
        
        # Grid
        ax.grid(True)
        ax.tick_params( labelsize = self.fontsize )
        
        # Compact
        self.fig_tree_dyna.tight_layout()
        
        # Save ax
        self.ax_tree_dyna   = ax
        self.node_wait_list = 0
        
        plt.ion()
        
        
     ############################
    def dyna_plot_add_node(self, node , no_nodes ):
        
        if not(node.parent==None):
                self.ax_tree_dyna.plot( 
                        [node.x[ self.x_axis ],node.parent.x[ self.x_axis ]] ,
                        [node.x[ self.y_axis ],node.parent.x[ self.y_axis ]] ,
                        'o-')
                self.time_text.set_text(self.time_template % ( no_nodes ))
                self.node_wait_list = self.node_wait_list + 1
                
                # Update graph
                if self.node_wait_list == self.dyna_node_no_update:
                    plt.pause( 0.001 )
                    self.node_wait_list = 0
                    print(
                    '\nRRT searching ( total node number = ', no_nodes, ' )'
                    )
                    
                    
    ############################
    def dyna_plot_clear(self ):
        
        self.ax_tree_dyna.clear()
        plt.close( self.fig_tree_dyna )
        self.dyna_plot_init()
                    
                    
    ############################
    def dyna_plot_solution(self ):
        
        if not self.solution_is_found == None:
            for node in self.path_node_list:
                if not(node.parent==None):
                    self.ax_tree_dyna.plot( 
                        [node.x[ self.x_axis ],node.parent.x[ self.x_axis ]] ,
                        [node.x[ self.y_axis ],node.parent.x[ self.y_axis ]] ,
                        'r')
                    
            #plt.ioff()
            self.fig_tree_dyna.show()
            
            

'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    from pyro.dynamic import pendulum
    

    sys  = pendulum.SinglePendulum()
    
    
    
    planner = RRT( sys )
    
    planner.x_start = np.array([ 0.1   , 0 ])
    planner.x_goal  = np.array([ -3.14 , 0 ])
    
    planner.u_options = [
            np.array([-5]),
            np.array([-3]),
            np.array([-1]),
            np.array([ 0]),
            np.array([ 1]),
            np.array([ 3]),
            np.array([ 5])
            ]
    
    planner.goal_radius = 0.6
    planner.alpha = 0.99
    
    
    planner.compute_solution()
    
    planner.plot_tree()
    planner.show_solution()
    planner.low_pass_solution( 3 )
    planner.show_solution()
    planner.z_axis = 0
    planner.plot_tree_3d()
    planner.animate_solution()


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

from pyro.dynamic import massspringdamper

sys = massspringdamper.ThreeMass()

sys.animate_eigen_mode(0)
sys.animate_eigen_mode(2)
sys.animate_eigen_mode(4)


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:17:36 2018

@author: Alexandre
"""

##############################################################################
import numpy as np

##############################################################################
from pyro.dynamic  import system
from pyro.control  import controller
##############################################################################

##############################################################################

class DoubleIntegrator2( system.ContinuousDynamicSystem ):

    
    ############################
    def __init__(self):
        """ """
        
        # Dimensions
        self.n = 2   
        self.m = 1   
        self.p = 2
        
        # initialize standard params
        super().__init__(self.n, self.m, self.p)
    
        # Labels
        self.name = 'Double Integrator'
        self.state_label = ['Position','Speed']
        self.input_label = ['Force']
        self.output_label = ['Position']
        
        # Units
        self.state_units = ['[m]','[m/sec]']
        self.input_units = ['[N]']
        self.output_units = ['[m]']
        
    
    #############################
    def f(self, x = np.zeros(2) , u = np.zeros(1) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        # Example double intergrator
        # x[0]: position x[1]: speed
        
        dx[0] = x[1]  # 
        dx[1] = u[0]  # 
        
        return dx


################
# Controller
################
    
    
class OptmimalDoubleIntegratorController( controller.StaticController ) :
    
    ############################
    def __init__( self ):
        """ """
        
        # Dimensions
        self.dof = 1
        self.k   = 1 
        self.m   = 1
        self.p   = 2
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Optmimal Double Integrator Controller'
        
        self.xd = 0

        self.rbar = np.zeros(1)
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(1) 
        
        x  = y[0]
        dx = y[1]
        
        xe = x - self.xd
        
        if (dx > np.sign(-xe) * np.sqrt(np.abs(2 * ( xe )))) :
            u[0] = -1
        else:
            u[0] = 1

        
        return u

    
###################################
# DOuble integrator
###################################

sys = DoubleIntegrator2()


ctl = OptmimalDoubleIntegratorController()

ctl.xd = 2

# New cl-dynamic
cl_sys = ctl + sys

cl_sys.x0 = np.array([-2,-2])
cl_sys.compute_trajectory( 10 , 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

from pyro.dynamic import massspringdamper

sys = massspringdamper.TwoMass()

sys.m1 = 1
sys.m2 = 1
sys.k1 = 1
sys.k2 = 1
sys.b1 = 0
sys.b2 = 0

sys.compute_ABCD()


sys.animate_eigen_mode(0)
sys.animate_eigen_mode(2)


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

# Target
q_desired = np.array([0.5,0.5])

# Joint pd

dof = 2

joint_pd      = robotcontrollers.JointPD( dof )
joint_pd.rbar = q_desired
joint_pd.kp   = np.array([25, 5 ])
joint_pd.kd   = np.array([ 1, 0 ])


# Closed-loops

robot_with_joint_pd    = joint_pd + torque_controlled_robot 

# Simulations
tf = 4
robot_with_joint_pd.x0 = np.array([0,0,0,0])
robot_with_joint_pd.compute_trajectory( tf )
robot_with_joint_pd.plot_trajectory('xu')
robot_with_joint_pd.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

from pyro.dynamic import massspringdamper

sys = massspringdamper.ThreeMass()


sys.plot_linearized_bode()
sys.plot_linearized_pz_map()

sys.x0[2] = 1

sys.plot_trajectory()
sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct  3 08:27:06 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.vehicle_steering                 import KinematicCarModel
from pyro.planning.trajectoryoptimisation import DirectCollocationTrajectoryOptimisation


sys = KinematicCarModel()

# Constraints
sys.x_ub = np.array([ 10, 2, 1.5])
sys.x_lb = np.array([-10,-1,-1.5])
sys.u_ub = np.array([ 10, 0.5])
sys.u_lb = np.array([-10, -0.5])

# Cost function
sys.cost_function.Q[0,0] = 1
sys.cost_function.Q[1,1] = 1
sys.cost_function.Q[2,2] = 10
sys.cost_function.R[0,0] = 1
sys.cost_function.R[1,1] = 1000

planner = DirectCollocationTrajectoryOptimisation( sys , 0.1, 20)

#PArking
planner.x_start = np.array([-0,1,0])
planner.x_goal  = np.array([ 0,0,0])

planner.maxiter = 1000
# planner.ini
planner.compute_optimal_trajectory()
# planner.show_solution()
planner.animate_solution()




# -*- coding: utf-8 -*-
"""
Created on Thu Oct 18 20:54:31 2018

@author: Alexandre
"""

import numpy as np

from pyro.dynamic import system


##############################################################################
        
class SimpleIntegrator( system.ContinuousDynamicSystem ):
    """ 
    SimpleIntegrator Example for a ContinuousDynamicSystem
    -------------------------------------------------------
    mass driven by a speed input: dx [m/sec] = f(x,u,t) = u [m/sec]
    
    """
    
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 1   
        self.m = 1   
        self.p = 1
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__( self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Simple Integrator'
        self.state_label = ['Position']
        self.input_label = ['Speed']
        self.output_label = ['Position']
        
        # Units
        self.state_units = ['[m]']
        self.input_units = ['[m/sec]']
        self.output_units = ['[m]']
        
    
    #############################
    def f(self, x = np.zeros(2) , u = np.zeros(1) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        # Example simpleintergrator
        # x[0]: position u[0]: speed
        
        dx[0] = u
        
        return dx


#############################################################################

class DoubleIntegrator( system.ContinuousDynamicSystem ):
    """ 
    DoubleIntegrator Example for a ContinuousDynamicSystem
    
    """
    
    ############################
    def __init__(self):
        """ """
        
        # Dimensions
        self.n = 2   
        self.m = 1   
        self.p = 1
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__( self, self.n, self.m, self.p)
    
        # Labels
        self.name = 'Double Integrator'
        self.state_label = ['Position','Speed']
        self.input_label = ['Force']
        self.output_label = ['Position']
        
        # Units
        self.state_units = ['[m]','[m/sec]']
        self.input_units = ['[N]']
        self.output_units = ['[m]']
        
    
    #############################
    def f(self, x = np.zeros(2) , u = np.zeros(1) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        # Example double intergrator
        # x[0]: position x[1]: speed
        
        dx[0] = x[1]  # 
        dx[1] = u[0]  # 
        
        return dx
    
    
    #############################
    def h( self , x , u , t ):
        """ 
        Output fonction y = h(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUTPUTS
        y  : output derivative vector p x 1
        
        """
        
        y = np.zeros(self.p) # Output vector
        
        y[0] = x[0] # output is first state = position
        
        return y
    
    
    
##############################################################################
        
    
class TripleIntegrator( system.ContinuousDynamicSystem ):
    """ 
    DoubleIntegrator Example for a ContinuousDynamicSystem
    
    """
    ############################
    def __init__(self):
        """ """
    
        # Dimensions
        self.n = 3   
        self.m = 1   
        self.p = 1
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__( self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Triple Integrator'
        self.state_label = ['Position','Speed', 'Force']
        self.input_label = ['Gradient of Force']
        self.output_label = ['Position']
        
        # Units
        self.state_units = ['[m]','[m/sec]','[N]']
        self.input_units = ['[N/sec]']
        self.output_units = ['[m]']
        
    
    #############################
    def f(self, x = np.zeros(2) , u = np.zeros(1) , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        # x[0]: position 
        # x[1]: speed 
        # x[2]: force
        
        dx[0] = x[1]  # 
        dx[1] = x[2]  # 
        dx[2] = u[0]  # 
        
        return dx
    
    
    #############################
    def h( self , x , u , t ):
        """ 
        Output fonction y = h(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUTPUTS
        y  : output derivative vector o x 1
        
        """
        
        y = np.zeros(self.p) # Output vector
        
        y[0] = x[0] # output is first state = position
        
        return y
    
    
    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    
    from pyro.analysis import costfunction
    
    ###################################
    # Simple integrator
    ###################################
    
    si = SimpleIntegrator()
    
    # Phase plane
    si.ubar = np.array([1])
    si.plot_phase_plane(0,0) # only one state for two axis!
    
    # Simulation
    si.x0 = np.array([2])
    si.plot_trajectory('xuj')
    si.plot_phase_plane_trajectory(0,0)
    
    ###################################
    # Double integrator
    ###################################
    
    di = DoubleIntegrator()
    
    # Phase plane behavior test
    di.ubar = np.array([1])   
    di.plot_phase_plane()
    
    # Simulation
    di.x0 = np.array([0,0])
    di.plot_trajectory('xuj')
    di.plot_trajectory('y')
    
    # Time cost
    cost_function =  costfunction.TimeCostFunction( di.xbar )
    di.traj       =  cost_function.trajectory_evaluation( di.traj )
    di.plot_trajectory('xuj')
    
    # Phase plane trajectory
    di.plot_phase_plane_trajectory()

    ###################################
    # Simple integrator
    ###################################
    
    ti = TripleIntegrator()
    
    # Phase plane
    ti.ubar = np.array([1])
    
    # Simulation
    ti.x0 = np.array([2,0,0]) 
    ti.plot_trajectory()
    ti.plot_phase_plane_trajectory_3d()